from datetime import timedelta
from functools import partial
from typing import Callable, ClassVar

from hyperopt import fmin, space_eval, tpe

from fedot.core.optimisers.objective import PipelineObjectiveEvaluate
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.search_space import SearchSpace, convert_params
from fedot.core.pipelines.tuning.tuner_interface import HyperoptTuner


class SequentialTuner(HyperoptTuner):
    """
    Class for hyperparameters optimization for all nodes sequentially
    """

    def __init__(self, objective_evaluate: PipelineObjectiveEvaluate,
                 iterations=100, early_stopping_rounds=None,
                 timeout: timedelta = timedelta(minutes=5),
                 inverse_node_order=False,
                 search_space: ClassVar = SearchSpace(),
                 algo: Callable = tpe.suggest,
                 n_jobs: int = -1):
        super().__init__(objective_evaluate, iterations, early_stopping_rounds, timeout, search_space, algo, n_jobs)

        self.inverse_node_order = inverse_node_order

    def tune(self, pipeline: Pipeline) -> Pipeline:
        """ Function for hyperparameters tuning on the entire pipeline

        :param pipeline: Pipeline which hyperparameters will be tuned
        """
        # Check source metrics for data
        self.init_check(pipeline)

        pipeline.replace_n_jobs_in_nodes(n_jobs=self.n_jobs)

        # Calculate amount of iterations we can apply per node
        nodes_amount = pipeline.length
        iterations_per_node = round(self.iterations / nodes_amount)
        iterations_per_node = int(iterations_per_node)
        if iterations_per_node == 0:
            iterations_per_node = 1

        # Calculate amount of seconds we can apply per node
        if self.max_seconds is not None:
            seconds_per_node = round(self.max_seconds / nodes_amount)
            seconds_per_node = int(seconds_per_node)
        else:
            seconds_per_node = None

        # Tuning performed sequentially for every node - so get ids of nodes
        nodes_ids = self.get_nodes_order(nodes_number=nodes_amount)
        for node_id in nodes_ids:
            node = pipeline.nodes[node_id]
            operation_name = node.operation.operation_type

            # Get node's parameters to optimize
            node_params = self.search_space.get_node_params(node_id=node_id,
                                                            operation_name=operation_name)

            if node_params is None:
                self.log.info(f'"{operation_name}" operation has no parameters to optimize')
            else:
                # Apply tuning for current node
                self._optimize_node(node_id=node_id,
                                    pipeline=pipeline,
                                    node_params=node_params,
                                    iterations_per_node=iterations_per_node,
                                    seconds_per_node=seconds_per_node)

        # Validation is the optimization do well
        final_pipeline = self.final_check(tuned_pipeline=pipeline)

        return final_pipeline

    def get_nodes_order(self, nodes_number: int) -> range:
        """ Method returns list with indices of nodes in the pipeline """

        if self.inverse_node_order is True:
            # From source data to output
            nodes_ids = range(nodes_number - 1, -1, -1)
        else:
            # From output to source data
            nodes_ids = range(0, nodes_number)

        return nodes_ids

    def tune_node(self, pipeline: Pipeline, node_index: int) -> Pipeline:
        """ Method for hyperparameters tuning for particular node

        :param pipeline: Pipeline which contains a node to be tuned
        :param node_index: Index of the node to tune
        """
        # Check source metrics for data
        self.init_check(pipeline)

        node = pipeline.nodes[node_index]
        operation_name = node.operation.operation_type

        # Get node's parameters to optimize
        node_params = self.search_space.get_node_params(node_id=node_index,
                                                        operation_name=operation_name)

        if node_params is None:
            self.log.info(f'"{operation_name}" operation has no parameters to optimize')
        else:
            # Apply tuning for current node
            self._optimize_node(pipeline=pipeline,
                                node_id=node_index,
                                node_params=node_params,
                                iterations_per_node=self.iterations,
                                seconds_per_node=self.max_seconds,
                                )

        # Validation is the optimization do well
        final_pipeline = self.final_check(tuned_pipeline=pipeline)
        return final_pipeline

    def _optimize_node(self, pipeline: Pipeline, node_id: int, node_params: dict, iterations_per_node: int,
                       seconds_per_node: int) -> Pipeline:
        """
        Method for node optimization

        :param pipeline: Pipeline which node is optimized
        :param node_id: id of the current node in the pipeline
        :param node_params: dictionary with parameters for node
        :param iterations_per_node: amount of iterations to produce
        :param seconds_per_node: amount of seconds to produce

        :return : updated pipeline with tuned parameters in particular node
        """
        best_parameters = fmin(partial(self._objective,
                                       pipeline=pipeline,
                                       node_id=node_id
                                       ),
                               node_params,
                               algo=self.algo,
                               max_evals=iterations_per_node,
                               early_stop_fn=self.early_stop_fn,
                               timeout=seconds_per_node)

        best_parameters = space_eval(space=node_params,
                                     hp_assignment=best_parameters)

        # Set best params for this node in the pipeline
        self.pipeline = self.set_arg_node(pipeline=pipeline,
                                          node_id=node_id,
                                          node_params=best_parameters)
        return self.pipeline

    def _objective(self, node_params: dict, pipeline: Pipeline, node_id: int) -> float:
        """
        Objective function for minimization / maximization problem

        :param node_params: dictionary with parameters for node
        :param pipeline: pipeline to evaluate
        :param node_id: id of the node to which parameters should be assigned

        :return metric_value: value of objective function
        """

        # Set hyperparameters for node
        pipeline = self.set_arg_node(pipeline=pipeline, node_id=node_id,
                                     node_params=node_params)

        metric_value = self.get_metric_value(pipeline=pipeline)
        return metric_value

    @staticmethod
    def set_arg_node(pipeline: Pipeline, node_id: int, node_params: dict) -> Pipeline:
        """ Method for parameters setting to a pipeline

        :param pipeline: pipeline which contains the node
        :param node_id: id of the node to which parameters should be assigned
        :param node_params: dictionary with labeled parameters to set

        :return pipeline: pipeline with new hyperparameters in each node
        """

        # Remove label prefixes
        node_params = convert_params(node_params)

        # Update parameters in nodes
        pipeline.nodes[node_id].parameters = node_params

        return pipeline



from copy import deepcopy

import numpy as np
from sklearn.metrics import roc_auc_score as roc_auc

from cases.data.data_utils import get_scoring_case_data_paths
from examples.simple.classification.classification_pipelines import classification_complex_pipeline
from fedot.core.data.data import InputData
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum


def get_case_train_test_data():
    """ Function for getting data for train and validation """
    train_file_path, test_file_path = get_scoring_case_data_paths()

    train_data = InputData.from_csv(train_file_path)
    test_data = InputData.from_csv(test_file_path)
    return train_data, test_data


def pipeline_tuning(pipeline: Pipeline, train_data: InputData,
                    test_data: InputData, local_iter: int,
                    tuner_iter_num: int = 30) -> (float, list):
    """ Function for tuning pipeline with PipelineTuner

    :param pipeline: pipeline to tune
    :param train_data: InputData for train
    :param test_data: InputData for validation
    :param local_iter: amount of tuner launches
    :param tuner_iter_num: amount of iterations, which tuner will perform

    :return mean_metric: mean value of ROC AUC metric
    :return several_iter_scores_test: list with metrics
    """
    several_iter_scores_test = []
    tuner = TunerBuilder(train_data.task) \
        .with_tuner(PipelineTuner) \
        .with_metric(ClassificationMetricsEnum.ROCAUC) \
        .with_iterations(tuner_iter_num) \
        .build(train_data)
    for iteration in range(local_iter):
        print(f'current local iteration {iteration}')

        # Pipeline tuning
        tuned_pipeline = tuner.tune(pipeline)

        # After tuning prediction
        tuned_pipeline.fit(train_data)
        after_tuning_predicted = tuned_pipeline.predict(test_data)

        # Metrics
        aft_tun_roc_auc = roc_auc(y_true=test_data.target,
                                  y_score=after_tuning_predicted.predict)
        several_iter_scores_test.append(aft_tun_roc_auc)

    max_metric = float(np.max(several_iter_scores_test))
    return max_metric, several_iter_scores_test


if __name__ == '__main__':
    train_data, test_data = get_case_train_test_data()

    # Pipeline composition
    pipeline = classification_complex_pipeline()

    # Before tuning prediction
    pipeline.fit(train_data)
    before_tuning_predicted = pipeline.predict(test_data)
    bfr_tun_roc_auc = roc_auc(y_true=test_data.target,
                              y_score=before_tuning_predicted.predict)

    local_iter = 5
    # Pipeline tuning
    after_tune_roc_auc, several_iter_scores_test = pipeline_tuning(pipeline=pipeline,
                                                                   train_data=train_data,
                                                                   test_data=test_data,
                                                                   local_iter=local_iter)

    print(f'Several test scores {several_iter_scores_test}')
    print(f'Maximal test score over {local_iter} iterations: {after_tune_roc_auc}')
    print(f'ROC-AUC before tuning {round(bfr_tun_roc_auc, 3)}')
    print(f'ROC-AUC after tuning {round(after_tune_roc_auc, 3)}')



from typing import Callable, Union, Sequence

from fedot.core.composer.metrics import Metric
from fedot.core.repository.quality_metrics_repository import (ClassificationMetricsEnum, ClusteringMetricsEnum,
                                                              ComplexityMetricsEnum, RegressionMetricsEnum)


class ApiMetrics:
    """
    Class for metrics matching. Handling both "metric name" - "metric instance"
    both for composer and tuner
    """

    _task_dict = {
        'regression': ['rmse', 'mae'],
        'classification': ['roc_auc', 'f1'],
        'multiclassification': 'f1',
        'clustering': 'silhouette',
        'ts_forecasting': ['rmse', 'mae']
    }

    _metrics_dict = {
        'acc': ClassificationMetricsEnum.accuracy,
        'roc_auc': ClassificationMetricsEnum.ROCAUC,
        'f1': ClassificationMetricsEnum.f1,
        'logloss': ClassificationMetricsEnum.logloss,
        'mae': RegressionMetricsEnum.MAE,
        'mse': RegressionMetricsEnum.MSE,
        'msle': RegressionMetricsEnum.MSLE,
        'mape': RegressionMetricsEnum.MAPE,
        'smape': RegressionMetricsEnum.SMAPE,
        'r2': RegressionMetricsEnum.R2,
        'rmse': RegressionMetricsEnum.RMSE,
        'rmse_pen': RegressionMetricsEnum.RMSE_penalty,
        'silhouette': ClusteringMetricsEnum.silhouette,
        'node_num': ComplexityMetricsEnum.node_num
    }

    def __init__(self, problem: str):
        if '/' in problem:
            # Solve multitask problem
            self.main_problem, self.side_problem = problem.split('/')
        else:
            self.main_problem = problem
            self.side_problem = None

    def get_problem_metrics(self) -> Union[str, Sequence[str]]:
        return ApiMetrics._task_dict[self.main_problem]

    @staticmethod
    def get_metrics_mapping(metric_name: Union[str, Callable]) -> Union[Metric, Callable]:
        if isinstance(metric_name, Callable):
            # for custom metric
            return metric_name
        return ApiMetrics._metrics_dict[metric_name]



from typing import Optional

import numpy as np
from sklearn.feature_selection import RFE
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from fedot.core.data.data import OutputData, InputData
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import \
    DataOperationImplementation
from fedot.core.operations.operation_parameters import OperationParameters


class FeatureSelectionImplementation(DataOperationImplementation):
    """ Class for applying feature selection operations on tabular data """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = None
        self.operation = None
        self.is_not_fitted = None

        # Number of columns in features table
        self.features_columns_number = None

        # Bool mask where True - remain column and False - drop it
        self.remain_features_mask = None

    def fit(self, input_data: InputData):
        """ Method for fit feature selection

        :param input_data: data with features, target and ids to process
        :return operation: trained operation (optional output)
        """
        features = input_data.features
        target = input_data.target

        # Define number of columns in the features table
        if len(features.shape) == 1:
            self.features_columns_number = 1
        else:
            self.features_columns_number = features.shape[1]

        if self.features_columns_number > 1:
            if self._is_input_data_one_dimensional(features):
                self.is_not_fitted = True
                return self.operation
            try:
                self.operation.fit(features, target)
            except ValueError:
                # For time series forecasting not available multi-targets
                self.operation.fit(features, target[:, 0])
        else:
            self.is_not_fitted = True
        return self.operation

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for making prediction for prediction stage

        :param input_data: data with features, target and ids to process
        :return output_data: filtered input data by columns
        """
        if self.is_not_fitted:
            return self._convert_to_output(input_data, input_data.features)

        features = input_data.features
        source_features_shape = features.shape
        transformed_features = self._make_new_table(features)

        # Update features
        output_data = self._convert_to_output(input_data,
                                              transformed_features)
        self._update_column_types(source_features_shape, output_data)
        return output_data

    def _update_column_types(self, source_features_shape, output_data: OutputData):
        """ Update column types after applying feature selection operations """
        if len(source_features_shape) < 2:
            return output_data
        else:
            if self.features_columns_number > 1:
                cols_number_removed = source_features_shape[1] - output_data.predict.shape[1]
                if cols_number_removed > 0:
                    # There are several columns, which were dropped
                    col_types = output_data.supplementary_data.column_types['features']

                    # Calculate
                    remained_column_types = np.array(col_types)[self.remain_features_mask]
                    output_data.supplementary_data.column_types['features'] = list(remained_column_types)

    def _make_new_table(self, features):
        """
        The method creates a table based on transformed data and source boolean
        features

        :param features: tabular data for processing
        :return transformed_features: transformed features table
        """

        # Bool vector - mask for columns
        self.remain_features_mask = self.operation.support_
        transformed_features = features[:, self.remain_features_mask]
        return transformed_features

    @staticmethod
    def _is_input_data_one_dimensional(features_to_process: np.array):
        """ Check if features table contain only one column """
        return features_to_process.shape[1] == 1


class LinearRegFSImplementation(FeatureSelectionImplementation):
    """
    Class for feature selection based on Recursive Feature Elimination (RFE) and
    LinearRegression as core model
    Task type - regression
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = LinearRegression()

        if not self.params:
            # Default parameters
            self.operation = RFE(estimator=self.inner_model)
        else:
            # Checking the appropriate params are using or not
            rfe_params = {k: self.params.get(k) for k in
                          ['n_features_to_select', 'step']}
            self.operation = RFE(estimator=self.inner_model, **rfe_params)


class NonLinearRegFSImplementation(FeatureSelectionImplementation):
    """
    Class for feature selection based on Recursive Feature Elimination (RFE) and
    DecisionTreeRegressor as core model
    Task type - regression
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = DecisionTreeRegressor()

        if not self.params:
            # Default parameters
            self.operation = RFE(estimator=self.inner_model)
        else:
            # Checking the appropriate params are using or not
            rfe_params = {k: self.params.get(k) for k in
                          ['n_features_to_select', 'step']}
            self.operation = RFE(estimator=self.inner_model, **rfe_params)


class LinearClassFSImplementation(FeatureSelectionImplementation):
    """
    Class for feature selection based on Recursive Feature Elimination (RFE) and
    LogisticRegression as core model
    Task type - classification
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = LogisticRegression()

        if not self.params:
            # Default parameters
            self.operation = RFE(estimator=self.inner_model)
        else:
            # Checking the appropriate params are using or not
            rfe_params = {k: self.params.get(k) for k in
                          ['n_features_to_select', 'step']}
            self.operation = RFE(estimator=self.inner_model, **rfe_params)


class NonLinearClassFSImplementation(FeatureSelectionImplementation):
    """
    Class for feature selection based on Recursive Feature Elimination (RFE) and
    DecisionTreeClassifier as core model
    Task type - classification
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.inner_model = DecisionTreeClassifier()

        if not self.params:
            # Default parameters
            self.operation = RFE(estimator=self.inner_model)
        else:
            # Checking the appropriate params are using or not
            rfe_params = {k: self.params.get(k) for k in
                          ['n_features_to_select', 'step']}
            self.operation = RFE(estimator=self.inner_model, **rfe_params)



import json
import os
from collections import Counter
from datetime import datetime
from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import joblib
import numpy as np

from fedot.core.log import default_log
from fedot.core.operations.atomized_template import AtomizedModelTemplate
from fedot.core.operations.operation_template import OperationTemplate, check_existing_path
from fedot.core.pipelines.node import Node, PrimaryNode, SecondaryNode

if TYPE_CHECKING:
    from fedot.core.pipelines.pipeline import Pipeline

from fedot.core.repository.operation_types_repository import atomized_model_type


class NumpyIntEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        return super().default(obj)


class PipelineTemplate:
    """
    Pipeline wrapper with 'export_pipeline'/'import_pipeline' methods
    allowing user to upload a pipeline to JSON format and import it from JSON.

    :params pipeline: Pipeline object to export or empty Pipeline to import
    """

    def __init__(self, pipeline: 'Pipeline' = None):
        self.total_pipeline_operations = Counter()
        self.operation_templates: List[OperationTemplate] = []
        self.unique_pipeline_id = str(uuid4())
        self.metadata: Dict[str, Any] = {}
        if pipeline is not None:
            self.depth = pipeline.depth
            self.metadata['computation_time_in_seconds'] = pipeline.computation_time

            # Save preprocessing operations
            self.data_preprocessor = pipeline.preprocessor
        else:
            self.depth = 0
            self.data_preprocessor = None

        self.log = default_log(self)

        self._pipeline_to_template(pipeline)

    def _pipeline_to_template(self, pipeline):
        try:
            # TODO improve for graph with several roots
            self._extract_pipeline_structure(pipeline.root_node, 0, [])
        except Exception as ex:
            self.log.info(f'Cannot export to template: {ex}')
        self.link_to_empty_pipeline = pipeline

    def _extract_pipeline_structure(self, node: Node, operation_id: int, visited_nodes: List[Node]):
        """
        Recursively go through the Pipeline from 'root_node' to PrimaryNode's,
        creating a OperationTemplate with unique id for each Node. In addition,
        checking whether this Node has been visited yet.
        """
        if node and node.nodes_from:
            nodes_from = []
            for node_parent in node.nodes_from:
                if node_parent in visited_nodes:
                    nodes_from.append(visited_nodes.index(node_parent) + 1)
                else:
                    visited_nodes.append(node_parent)
                    nodes_from.append(len(visited_nodes))
                    self._extract_pipeline_structure(node_parent, len(visited_nodes), visited_nodes)
        else:
            nodes_from = []

        # TODO resolve as to GraphTemplate
        if hasattr(node, 'operation'):
            if (not isinstance(node.operation, str) and
                    node.operation.operation_type == atomized_model_type()):
                operation_template = AtomizedModelTemplate(node, operation_id, sorted(nodes_from))
            else:
                operation_template = OperationTemplate(node, operation_id, sorted(nodes_from))

            self.operation_templates.append(operation_template)
            self.total_pipeline_operations[operation_template.operation_type] += 1

    def export_pipeline(self, path: Optional[str] = None, root_node: Optional[Node] = None,
                        create_subdir: bool = True,
                        is_datetime_in_path: bool = False) -> Tuple[str, dict]:
        """
        Save JSON to path and return this JSON like object

        Args:
            path: custom path to save JSON to
            root_node: root node of the exported pipeline
            create_subdir: if True -- create one more dir in the last dir.
                           Useful for saving pipelines with fitted_operations & preprocessing.
                           if False -- save to the last dir in this path.
                           Useful for pipelines without fitted_operations or
                           when you want to know the exact name of the folder.
            is_datetime_in_path: is it required to add the datetime timestamp to the path.

        :return: <JSON representation of the pipeline structure>, <dict of paths to fitted models>
        """

        pipeline_template_dict = self.convert_to_dict(root_node)
        fitted_ops = {}
        if path is None:
            fitted_ops = self._create_fitted_operations()
            if fitted_ops is not None:
                for operation in pipeline_template_dict['nodes']:
                    saved_key = f'operation_{operation["operation_id"]}'
                    if saved_key not in fitted_ops:
                        saved_key = None
                    pipeline_template_dict['fitted_operation_path'] = saved_key

        json_data = json.dumps(pipeline_template_dict, indent=4, cls=NumpyIntEncoder)

        if path is None:
            return json_data, fitted_ops

        path_to_dir, path_to_pipe = self._prepare_paths(path, create_subdir=create_subdir,
                                                        is_datetime_in_path=is_datetime_in_path)
        with open(path_to_pipe, 'w', encoding='utf-8') as f:
            f.write(json_data)
            self.log.debug(f'The pipeline saved in the path: {path_to_pipe}.')

        dict_fitted_operations = self._create_fitted_operations(path_to_dir)

        return json_data, dict_fitted_operations

    def convert_to_dict(self, root_node: Node = None) -> dict:
        """ Generate pipeline description in a form of dictionary """

        json_nodes = list(map(lambda op_template: op_template.convert_to_dict(), self.operation_templates))
        for node in json_nodes:
            if 'custom_params' in node and isinstance(node['custom_params'], dict):
                for key in node['custom_params']:
                    if isinstance(node['custom_params'][key], Callable):
                        node['custom_params'][key] = None

        # Store information about preprocessing
        preprocessing_path = ['preprocessing', 'data_preprocessor.pkl']

        json_object = {
            "total_pipeline_operations": list(self.total_pipeline_operations),
            "depth": self.depth,
            "nodes": json_nodes,
            "preprocessing": preprocessing_path
        }
        if root_node:
            json_object['descriptive_id'] = root_node.descriptive_id

        return json_object

    def _create_fitted_operations(self, path=None):
        """ Create .pkl files for operations using absolute path """
        dict_fitted_operations = {}
        for operation in self.operation_templates:
            dict_fitted_operations[f'operation_{operation.operation_id}'] = operation.export_operation(path)

        if all(val is None for val in dict_fitted_operations.values()):
            return None

        # Save preprocessing module
        preprocessing_path = self.export_preprocessing(path)
        if isinstance(preprocessing_path, str):
            preprocessing_splitted = os.path.split(preprocessing_path)
            preprocessing_path = [preprocessing_splitted[-2], preprocessing_splitted[-1]]
        dict_fitted_operations['preprocessing'] = preprocessing_path
        return dict_fitted_operations

    @staticmethod
    def _prepare_paths(path: str, create_subdir: bool, is_datetime_in_path: bool = False) -> Tuple[str, str]:
        """ Constructs paths to pipeline folder and pipeline json
        Args:
            path: path to json object or to the folder
            create_subdir: bool indicating whether to save pipeline in specified dir or create additional subdir
            is_datetime_in_path: bool indicating whether to save with timestamp or not

        :return: paths to pipeline dir and to pipeline json
        """

        path = os.path.abspath(path)

        # explicitly specify where to save json, fitted_operation and etc will be saved in the same dir
        if path.endswith('.json'):
            path_to_dir = os.path.split(path)[0]
            if not os.path.exists(path_to_dir):
                os.makedirs(path_to_dir)
            path_to_json = path
            return path_to_dir, path_to_json

        if not os.path.exists(path):
            os.makedirs(path)

        suffix = '_pipeline_saved'

        if not create_subdir:
            path_to_dir = path
            folder_name = os.path.split(path_to_dir)[-1]
            if is_datetime_in_path:
                json_name = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}_{folder_name}"
            else:
                json_name = folder_name
            path_to_json = os.path.join(path_to_dir, f'{json_name}.json')
            return path_to_dir, path_to_json

        if is_datetime_in_path:
            folder_name = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}{suffix}"
        else:
            # get index for pipeline
            files_by_path = os.listdir(path)
            max_pipeline_idx = -1
            for file in files_by_path:
                if suffix in file:
                    cur_pipeline_idx = file.split(suffix)[0] \
                        if '_' not in file.split(suffix)[0] else 0
                    if cur_pipeline_idx.isdigit():
                        cur_pipeline_idx = int(cur_pipeline_idx)
                    else:
                        continue
                    max_pipeline_idx = cur_pipeline_idx if cur_pipeline_idx > max_pipeline_idx else max_pipeline_idx
            folder_name = f"{max_pipeline_idx+1}{suffix}"

        path_to_dir = os.path.join(path, folder_name)
        path_to_dir = path_to_dir
        if not os.path.exists(path_to_dir):
            os.makedirs(path_to_dir)

        folder_name = os.path.split(path_to_dir)[-1]
        path_to_json = os.path.join(path_to_dir, f'{folder_name}.json')

        return path_to_dir, path_to_json

    @staticmethod
    def _is_nested_path(path):
        return path.find('nested') == -1

    def import_pipeline(self, source: Union[str, dict], dict_fitted_operations: Optional[dict] = None):
        """
        Imports pipeline from source into the :attr:`link_to_empty_pipeline`

        :param source: where to load the pipeline from: can be specified as dict or as str
        :param dict_fitted_operations: dictionary of the fitted operations
        """
        path_to_json_pipe = None

        if source is None:
            raise ValueError('Cannot import pipeline: the source is None')
        elif isinstance(source, dict):
            json_object_pipeline = source
            self.log.debug('The pipeline was imported from dict.')
        else:
            path = os.path.abspath(source)
            if os.path.isdir(path):
                path_to_json_pipe = None
                for file in os.listdir(path):
                    if file.endswith('.json'):
                        path_to_json_pipe = os.path.join(path, file)
                if not path_to_json_pipe:
                    raise FileNotFoundError('Path to pipeline is incorrect')
            else:
                path_to_json_pipe = path

            with open(path_to_json_pipe) as json_file:
                json_object_pipeline = json.load(json_file)
                self.log.debug(f'The pipeline was imported from the path: {path_to_json_pipe}.')

        self._extract_operations(json_object_pipeline, path_to_json_pipe)
        self.convert_to_pipeline(self.link_to_empty_pipeline, path_to_json_pipe, dict_fitted_operations)
        self.depth = self.link_to_empty_pipeline.depth

    def _extract_operations(self, pipeline_json: dict, path: str):
        operation_objects = pipeline_json['nodes']

        # Update info about fitted operation
        for operation_object in operation_objects:
            if operation_object['operation_type'] == atomized_model_type():
                filename = operation_object['atomized_model_json_path'] + '.json'
                curr_path = os.path.join(os.path.dirname(path), operation_object['atomized_model_json_path'], filename)
                operation_template = AtomizedModelTemplate(path=curr_path)
            else:
                operation_template = OperationTemplate()

            operation_template.import_json(operation_object)
            self.operation_templates.append(operation_template)
            self.total_pipeline_operations[operation_template.operation_type] += 1

    def convert_to_pipeline(self, pipeline, path: str = None, dict_fitted_operations: dict = None):
        if path is not None:
            path = os.path.abspath(os.path.dirname(path))
        visited_nodes = {}
        root_template = [op_template for op_template in self.operation_templates if op_template.operation_id == 0][0]

        root_node = self.roll_pipeline_structure(root_template, visited_nodes, path, dict_fitted_operations)
        pipeline.nodes.clear()
        pipeline.add_node(root_node)

        preprocessor_file = None
        if path is not None and 'preprocessing' in os.listdir(path):
            # Load data preprocessor and store it into the
            preprocessor_file = os.path.join(path, 'preprocessing', 'data_preprocessor.pkl')
        elif dict_fitted_operations and 'preprocessing' in dict_fitted_operations:
            preprocessor_file = dict_fitted_operations['preprocessing']
        if preprocessor_file:
            try:
                pipeline.preprocessor = joblib.load(preprocessor_file)
            except ModuleNotFoundError as ex:
                self.log.warning(f'Could not load preprocessor from file `{preprocessor_file}` '
                                 f'due to legacy incompatibility. Please refit the preprocessor.')

    def roll_pipeline_structure(self, operation_object: Union['OperationTemplate', 'AtomizedModelTemplate'],
                                visited_nodes: dict, path: str = None, dict_fitted_operations: dict = None):
        """
        The function recursively traverses all disjoint operations
        and connects the operations in a pipeline.

        :params operation_object: operationTemplate or AtomizedOperationTemplate
        :params visited_nodes: array to remember which node was visited
        :params path: path to save
        :return: root_node
        """
        fitted_operation = None
        if operation_object.operation_id in visited_nodes:
            return visited_nodes[operation_object.operation_id]

        if operation_object.operation_type == atomized_model_type():
            atomized_model = operation_object.next_pipeline_template
            if operation_object.nodes_from:
                node = SecondaryNode(operation_type=atomized_model)
            else:
                node = PrimaryNode(operation_type=atomized_model)
        else:
            if operation_object.nodes_from:
                node = SecondaryNode(operation_object.operation_type)
            else:
                node = PrimaryNode(operation_object.operation_type)

            node.parameters = operation_object.custom_params

            node.rating = operation_object.rating

        if hasattr(operation_object,
                   'fitted_operation_path') and operation_object.fitted_operation_path and path is not None:
            path_to_operation = os.path.join(path, operation_object.fitted_operation_path)

            if 'h2o' in operation_object.operation_type:
                fitted_operation = load_h2o(path, self.log)

            elif not os.path.isfile(path_to_operation):
                message = f"Fitted operation on the path: {path_to_operation} does not exist."
                self.log.error(message)
            else:
                fitted_operation = joblib.load(path_to_operation)
        elif dict_fitted_operations is not None:
            if 'h2o' in operation_object.operation_type:
                message = 'Loading h2o models from dict is not supported'
                self.log.error(message)
                raise TypeError(message)
            else:
                fitted_operation = joblib.load(dict_fitted_operations[f'operation_{operation_object.operation_id}'])

        operation_object.fitted_operation = fitted_operation
        node.fitted_operation = fitted_operation

        nodes_from = [operation_template for operation_template in self.operation_templates
                      if operation_template.operation_id in operation_object.nodes_from]

        node.nodes_from = [self.roll_pipeline_structure(node_from, visited_nodes, path, dict_fitted_operations) for
                           node_from in nodes_from]

        visited_nodes[operation_object.operation_id] = node
        return node

    def export_preprocessing(self, path: str = None):
        """ Save preprocessing operations in pkl files and store full paths into dictionary """
        if path:
            path_fitted_preprocessing = os.path.join(path, 'preprocessing')
            check_existing_path(path_fitted_preprocessing)

            data_preprocessor_path = os.path.join(path_fitted_preprocessing, 'data_preprocessor.pkl')
            if self.data_preprocessor is not None:
                joblib.dump(self.data_preprocessor, data_preprocessor_path)
                return data_preprocessor_path
        else:
            # dictionary with bytes of fitted operations
            if self.data_preprocessor:
                bytes_container = BytesIO()
                joblib.dump(self.data_preprocessor, bytes_container)
                return bytes_container


def extract_subtree_root(root_operation_id: int, pipeline_template: PipelineTemplate):
    root_node = [operation_template for operation_template in pipeline_template.operation_templates
                 if operation_template.operation_id == root_operation_id][0]
    root_node = pipeline_template.roll_pipeline_structure(root_node, {})

    return root_node


def load_h2o(path_to_operation, log):
    from fedot.core.operations.evaluation.automl import H2OSerializationWrapper
    try:
        return H2OSerializationWrapper.load_operation(path_to_operation)
    except EnvironmentError as e:
        message = f'Obtained type of H2O pipeline does not serializable: {e}'
        log.error(message)
        raise EnvironmentError(message)



import os
from typing import Union

from sklearn.metrics import f1_score as f1

from cases.dataset_preparation import unpack_archived_data
from fedot.api.main import Fedot
from fedot.core.data.data import InputData, OutputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utils import fedot_project_root


def calculate_validation_metric(valid: Union[InputData, MultiModalData], pred: OutputData) -> float:
    """
    Calculates F1 score for predicted data

    :param valid: dataclass with true target
    :param pred: dataclass with model's prediction
    """

    real = valid.target
    predicted = pred.predict

    err = f1(y_true=real,
             y_pred=predicted, average='micro')

    return round(err, 2)


def prepare_multi_modal_data(files_path: str, task: Task, images_size: tuple = (128, 128)) -> MultiModalData:
    """
    Imports data from 3 different sources (table, images and text)

    :param files_path: path to data
    :param task: task to solve
    :param images_size: the requested size in pixels, as a 2-tuple of (width, height)
    :return: MultiModalData object which contains table, text and image data
    """

    path = os.path.join(str(fedot_project_root()), files_path)
    # unpacking of data archive
    unpack_archived_data(path)
    # import of table data
    data_num = InputData.from_json_files(path, fields_to_use=['votes', 'rating'],
                                         label='genres', task=task, is_multilabel=True, shuffle=False)

    class_labels = data_num.target

    img_files_path = f'{files_path}/*.jpeg'
    img_path = os.path.join(str(fedot_project_root()), img_files_path)

    # import of image data
    data_img = InputData.from_image(images=img_path, labels=class_labels, task=task, target_size=images_size)
    # import of text data
    data_text = InputData.from_json_files(path, fields_to_use=['plot'],
                                          label='genres', task=task,
                                          data_type=DataTypesEnum.text, is_multilabel=True, shuffle=False)

    data = MultiModalData({
        'data_source_img': data_img,
        'data_source_table': data_num,
        'data_source_text': data_text
    })

    return data


def run_multi_modal_pipeline(files_path: str, visualization=False) -> float:
    task = Task(TaskTypesEnum.classification)
    images_size = (224, 224)

    data = prepare_multi_modal_data(files_path, task, images_size)

    fit_data, predict_data = train_test_data_setup(data, shuffle_flag=True, split_ratio=0.6)

    automl_model = Fedot(problem='classification', timeout=15)
    pipeline = automl_model.fit(features=fit_data,
                                target=fit_data.target)

    if visualization:
        pipeline.show()

    prediction = pipeline.predict(predict_data, output_mode='labels')

    err = calculate_validation_metric(predict_data, prediction)

    print(f'F1 micro for validation sample is {err}')

    return err


if __name__ == '__main__':
    run_multi_modal_pipeline(files_path='examples/data/multimodal', visualization=True)



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from fedot.api.main import Fedot
from fedot.core.repository.tasks import TsForecastingParams


def train_test_split(dataframe: pd.DataFrame, forecast_horizon: int):
    train_df = []
    test_df = []
    for ts in list(dataframe['label'].unique()):
        ts_dataframe = dataframe[dataframe['label'] == ts]
        ts_dataframe = ts_dataframe.sort_values(by='datetime')

        train_len = len(ts_dataframe) - forecast_horizon
        train_df.append(ts_dataframe.head(train_len))
        test_df.append(ts_dataframe.tail(forecast_horizon))

    train_df = pd.concat(train_df)
    test_df = pd.concat(test_df)
    return train_df, test_df


def plot_results(full_df: pd.DataFrame, target_column: int, forecast: np.array,
                 forecast_horizon: int):
    """ Display forecasted vs actual values """
    target_df = full_df[full_df['label'] == target_column]
    datetime_xs = target_df['datetime']
    plt.plot(datetime_xs, target_df['value'], label='Actual values')
    plt.plot(datetime_xs.tail(forecast_horizon), forecast[:-forecast_horizon], label='Forecast')
    plt.xlabel('Datetime', fontsize=13)
    plt.ylabel('Sea surface height, m', fontsize=13)
    plt.legend()
    plt.grid()
    plt.show()


def launch_fedot_forecasting(target_column: int = 1, forecast_horizon: int = 50,
                             number_of_series_to_use: int = 25):
    """ Example how to launch FEDOT AutmoML for multivariate forecasting """
    path_to_file = os.path.join(os.path.curdir, 'data', 'multivariate_ssh.csv')
    df = pd.read_csv(path_to_file, parse_dates=['datetime'])
    train_df, test_df = train_test_split(df, forecast_horizon)

    # Prepare train data in a form of dictionary {'time series label': numpy array}
    train_data = {}
    feature_series = list(train_df['label'].unique())
    for feature_ts in feature_series[:number_of_series_to_use]:
        current_ts = train_df[train_df['label'] == feature_ts]
        train_data.update({str(feature_ts): np.array(current_ts['value'])})

    # Configure AutoML
    task_parameters = TsForecastingParams(forecast_length=forecast_horizon)
    model = Fedot(problem='ts_forecasting', task_params=task_parameters, timeout=3,
                  cv_folds=2, validation_blocks=2, with_tuning=False)
    target_series = train_df[train_df['label'] == target_column]
    obtained_pipeline = model.fit(features=train_data,
                                  target=np.array(target_series['value']))

    obtained_pipeline.show()

    # Use historical value to make forecast
    forecast = model.predict(train_data)

    plot_results(df, target_column, forecast, forecast_horizon)


if __name__ == '__main__':
    launch_fedot_forecasting(number_of_series_to_use=30)



import random
from typing import Any, Dict, List, Optional, Union

import numpy as np

from fedot.api.api_utils.presets import OperationsPreset
from fedot.core.constants import AUTO_PRESET_NAME, DEFAULT_FORECAST_LENGTH
from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.log import Log, default_log
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskParams, TaskTypesEnum, TsForecastingParams
from fedot.core.utilities.random import RandomStateHandler


class ApiParams:

    def __init__(self):
        self.api_params = None
        self.log = None
        self.task = None
        self.metric_to_compose = None
        self.metric_to_tuning = None
        self.task_params = None
        self.metrics_name = None

    def initialize_params(self, input_params: Dict[str, Any]):
        """ Merge input_params dictionary with several parameters for AutoML algorithm """
        self.get_initial_params(input_params)

        # Final check for correctness for timeout and generations
        self.api_params = check_timeout_vs_generations(self.api_params)

    def update_available_operations_by_preset(self, data: InputData):
        preset_operations = OperationsPreset(task=self.task, preset_name=self.api_params['preset'])
        self.api_params = preset_operations.composer_params_based_on_preset(self.api_params, data.data_type)

    def get_initial_params(self, input_params: Dict[str, Any]):
        self._parse_input_params(input_params)

        param_dict = {
            'task': self.task,
            'logger': self.log
        }
        self.api_params = {**self.api_params, **param_dict}

    def accept_and_apply_recommendations(self, input_data: Union[InputData, MultiModalData], recommendations: Dict):
        """
        Accepts recommendations for api params from DataAnalyser

        :param input_data - data for preprocessing
        :param recommendations - dict with recommendations
        """
        # TODO fix multimodality
        if isinstance(input_data, MultiModalData):
            self.api_params['cv_folds'] = None  # there are no support for multimodal data now
            for data_source_name, values in input_data.items():
                self.accept_and_apply_recommendations(input_data[data_source_name],
                                                      recommendations[data_source_name])
        else:
            if 'label_encoded' in recommendations:
                self.log.info("Change preset due to label encoding")
                self.change_preset_for_label_encoded_data(input_data.task, input_data.data_type)

    def change_preset_for_label_encoded_data(self, task: Task, data_type: DataTypesEnum):
        """ Change preset on tree like preset, if data had been label encoded """
        if 'preset' in self.api_params:
            preset_name = ''.join((self.api_params['preset'], '*tree'))
        else:
            preset_name = '*tree'
        preset_operations = OperationsPreset(task=task, preset_name=preset_name)

        if self.api_params.get('available_operations') is not None:
            del self.api_params['available_operations']
        self.api_params = preset_operations.composer_params_based_on_preset(self.api_params, data_type)
        param_dict = {
            'task': self.task,
            'logger': self.log
        }
        self.api_params = {**self.api_params, **param_dict}

    def _parse_input_params(self, input_params: Dict[str, Any]):
        """ Parses input params into different class fields """

        # reset logging level for Singleton
        Log().reset_logging_level(input_params['logging_level'])
        self.log = default_log(prefix='FEDOT logger')

        simple_keys = ['problem', 'n_jobs', 'timeout']
        self.api_params = {k: input_params[k] for k in simple_keys}

        default_evo_params = self.get_default_evo_params(self.api_params['problem'])
        if input_params['composer_tuner_params'] is None:
            evo_params = default_evo_params
        else:
            evo_params = {**default_evo_params, **input_params['composer_tuner_params']}
        self.api_params.update(evo_params)
        if 'preset' not in input_params['composer_tuner_params']:
            self.api_params['preset'] = 'auto'

        # If early_stopping_generations is not specified,
        # than estimate it as in time-based manner as: 0.33 * composing_timeout.
        # The minimal number of generations is 5.
        if 'early_stopping_iterations' not in input_params['composer_tuner_params']:
            if input_params['timeout']:
                depending_on_timeout = int(input_params['timeout']/3)
                self.api_params['early_stopping_iterations'] = \
                    depending_on_timeout if depending_on_timeout > 5 else 5

        specified_seed = input_params['seed']
        if specified_seed is not None:
            np.random.seed(specified_seed)
            random.seed(specified_seed)
            RandomStateHandler.MODEL_FITTING_SEED = specified_seed

        if self.api_params['problem'] == 'ts_forecasting' and input_params['task_params'] is None:
            self.log.warning(f'The value of the forecast depth was set to {DEFAULT_FORECAST_LENGTH}.')
            input_params['task_params'] = TsForecastingParams(forecast_length=DEFAULT_FORECAST_LENGTH)

        if self.api_params['problem'] == 'clustering':
            raise ValueError('This type of task is not not supported in API now')

        self.task = self.get_task_params(self.api_params['problem'], input_params['task_params'])
        self.metric_name = self.get_default_metric(self.api_params['problem'])
        self.api_params.pop('problem')

    @staticmethod
    def get_default_evo_params(problem: str):
        """ Dictionary with default parameters for composer """
        params = {'max_depth': 6,
                  'max_arity': 3,
                  'pop_size': 20,
                  'num_of_generations': None,
                  'keep_n_best': 1,
                  'with_tuning': True,
                  'preset': AUTO_PRESET_NAME,
                  'genetic_scheme': None,
                  'early_stopping_iterations': 30,
                  'early_stopping_timeout': 10,
                  'use_pipelines_cache': True,
                  'use_preprocessing_cache': True,
                  'cache_folder': None}

        if problem in ['classification', 'regression']:
            params['cv_folds'] = 5
        elif problem == 'ts_forecasting':
            params['cv_folds'] = 3
            params['validation_blocks'] = 2
        return params

    @staticmethod
    def get_default_metric(problem: str) -> Union[str, List[str]]:
        default_test_metric_dict = {
            'regression': ['rmse', 'mae'],
            'classification': ['roc_auc', 'f1'],
            'multiclassification': 'f1',
            'clustering': 'silhouette',
            'ts_forecasting': ['rmse', 'mae']
        }
        return default_test_metric_dict[problem]

    @staticmethod
    def get_task_params(problem: str, task_params: Optional[TaskParams] = None):
        """ Return task parameters by machine learning problem name (string) """
        task_dict = {'regression': Task(TaskTypesEnum.regression, task_params=task_params),
                     'classification': Task(TaskTypesEnum.classification, task_params=task_params),
                     'clustering': Task(TaskTypesEnum.clustering, task_params=task_params),
                     'ts_forecasting': Task(TaskTypesEnum.ts_forecasting, task_params=task_params)
                     }
        return task_dict[problem]


def check_timeout_vs_generations(api_params):
    timeout = api_params['timeout']
    num_of_generations = api_params['num_of_generations']
    if timeout in [-1, None]:
        api_params['timeout'] = None
        if num_of_generations is None:
            raise ValueError('"num_of_generations" should be specified if infinite "timeout" is given')
    elif timeout > 0:
        if num_of_generations is None:
            api_params['num_of_generations'] = 10000
    else:
        raise ValueError(f'invalid "timeout" value: timeout={timeout}')
    return api_params



import warnings
from typing import Optional

from sklearn.cluster import KMeans as SklearnKmeans

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.operations.evaluation.evaluation_interfaces import SkLearnEvaluationStrategy
from fedot.core.utilities.random import RandomStateHandler

warnings.filterwarnings("ignore", category=UserWarning)


class SkLearnClusteringStrategy(SkLearnEvaluationStrategy):
    __operations_by_types = {
        'kmeans': SklearnKmeans
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        if not params:
            self.params_for_fit.update(n_clusters=2)
        self.operation_impl = self._convert_to_operation(operation_type)

    def fit(self, train_data: InputData):
        """
        Fit method for clustering task

        :param train_data: data used for model training
        """

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        operation_implementation = self.operation_impl(**self.params_for_fit.to_dict())
        with RandomStateHandler():
            operation_implementation.fit(train_data.features)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Predict method for clustering task
        :param trained_operation: operation object
        :param predict_data: data used for prediction
        """
        prediction = trained_operation.predict(predict_data.features)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain SkLearn clustering strategy for {operation_type}')



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Optional

import numpy as np

from fedot.core.data.data import OutputData, InputData
from fedot.core.log import default_log
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum


class DataOperationImplementation(ABC):
    """ Interface for data operations realisations methods
    Contains abstract methods, which should be implemented for applying EA
    optimizer on it
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        self.params = params or OperationParameters()

    @abstractmethod
    def fit(self, input_data: InputData):
        """ Method fit operation on a dataset

        :param input_data: data with features, target and ids to process
        """
        raise NotImplementedError()

    @abstractmethod
    def transform(self, input_data: InputData) -> OutputData:
        """ Method apply transform operation on a dataset for predict stage

        :param input_data: data with features, target and ids to process
        """
        raise NotImplementedError()

    def transform_for_fit(self, input_data: InputData) -> OutputData:
        """ Method apply transform operation on a dataset for fit stage.
        Allows to implement transform method different from main transform method
        if another behaviour for fit graph stage is needed.

        :param input_data: data with features, target and ids to process
        """
        return self.transform(input_data)

    def get_params(self) -> OperationParameters:
        """ Method return parameters, which can be optimized for particular
        operation
        """
        return deepcopy(self.params)

    @staticmethod
    def _convert_to_output(input_data: InputData, predict: np.ndarray,
                           data_type: DataTypesEnum = DataTypesEnum.table) -> OutputData:
        """ Method prepare prediction of operation as OutputData object """

        converted = _convert_to_output_function(input_data, predict, data_type)
        return converted


class EncodedInvariantImplementation(DataOperationImplementation):
    """ Class for processing data without transforming encoded features.
    Encoded features - features after OneHot encoding operation, when one
    feature (with categorical values) can be represented as several boolean
    vectors
    """

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.operation = None
        self.ids_to_process = None
        self.bool_ids = None

    def fit(self, input_data: InputData):
        """ Method for fit transformer with automatic determination
        of boolean features, with which there is no need to make transformation

        :param input_data: data with features, target and ids to process
        :return operation: trained transformer (optional output)
        """

        features = input_data.features

        # Find boolean columns in features table
        bool_ids, ids_to_process = self._reasonability_check(features)
        self.ids_to_process = ids_to_process
        self.bool_ids = bool_ids

        if len(ids_to_process) > 0:
            features_to_process = np.array(features[:, ids_to_process])
            self.operation.fit(features_to_process)
        else:
            pass

        return self.operation

    def transform(self, input_data: InputData) -> OutputData:
        """
        The method that transforms the source features using "operation" for predict stage

        :param input_data: tabular data with features, target and ids to process
        :return output_data: output data with transformed features table
        """
        source_features_shape = input_data.features.shape
        features = input_data.features
        if len(self.ids_to_process) > 0:
            transformed_features = self._make_new_table(features)
        else:
            transformed_features = features

        # Update features and column types
        output_data = self._convert_to_output(input_data, transformed_features)
        self._update_column_types(source_features_shape, output_data)
        return output_data

    def _make_new_table(self, features):
        """
        The method creates a table based on transformed data and source boolean
        features

        :param features: tabular data for processing
        :return transformed_features: transformed features table
        """

        features_to_process = np.array(features[:, self.ids_to_process])
        transformed_part = self.operation.transform(features_to_process)

        # If there are no binary features in the dataset
        if len(self.bool_ids) == 0:
            transformed_features = transformed_part
        else:
            # Stack transformed features and bool features
            bool_features = np.array(features[:, self.bool_ids])
            frames = (bool_features, transformed_part)
            transformed_features = np.hstack(frames)

        return transformed_features

    def _update_column_types(self, source_features_shape, output_data: OutputData) -> OutputData:
        """
        Update column types after applying operations.
        If new columns added, new type for them are defined
        """
        return output_data

    @staticmethod
    def _reasonability_check(features):
        """
        Method for checking which columns contain boolean data

        :param features: tabular data for check
        :return bool_ids: indices of boolean columns in table
        :return non_bool_ids: indices of non boolean columns in table
        """
        # TODO perhaps there is a more effective way to do this
        source_shape = features.shape

        columns_amount = source_shape[1] if len(source_shape) > 1 else 1

        # Indices of boolean columns in features table
        bool_ids = []
        non_bool_ids = []

        # For every column in table make check
        for column_id in range(0, columns_amount):
            column = features[:, column_id] if columns_amount > 1 else features
            if len(np.unique(column)) > 2:
                non_bool_ids.append(column_id)
            else:
                bool_ids.append(column_id)

        return bool_ids, non_bool_ids


class ModelImplementation(ABC):
    """ Interface for models realisations methods
    Contains abstract methods, which should be implemented for applying EA
    optimizer on it
    """

    def __init__(self, params: OperationParameters = None):
        self.log = default_log(self)
        self.params = params or OperationParameters()

    @abstractmethod
    def fit(self, input_data: InputData):
        """ Method fit model on a dataset

        :param input_data: data with features, target and ids to process
        """
        raise NotImplementedError()

    @abstractmethod
    def predict(self, input_data: InputData) -> OutputData:
        """ Method make prediction

        :param input_data: data with features, target and ids to process
        """
        raise NotImplementedError()

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        """ Method make prediction while graph fitting.
        Allows to implement predict method different from main predict method
        if another behaviour for fit graph stage is needed.

        :param input_data: data with features, target and ids to process
        """
        return self.predict(input_data)

    def get_params(self) -> OperationParameters:
        """ Method return parameters, which can be optimized for particular
        operation
        """
        return deepcopy(self.params)

    @staticmethod
    def _convert_to_output(input_data: InputData, predict: np.array,
                           data_type: DataTypesEnum = DataTypesEnum.table):
        """ Method prepare prediction of operation as OutputData object """

        converted = _convert_to_output_function(input_data, predict, data_type)
        return converted


def _convert_to_output_function(input_data: InputData, transformed_features: np.array,
                                data_type: DataTypesEnum = DataTypesEnum.table):
    """ Function prepare prediction of operation as OutputData object

    :param input_data: data with features, target and ids to process
    :param transformed_features: transformed features
    :param data_type: type of output data
    """

    # After preprocessing operations by default we get tabular data
    converted = OutputData(idx=input_data.idx,
                           features=input_data.features,
                           predict=transformed_features,
                           task=input_data.task,
                           target=input_data.target,
                           data_type=data_type,
                           supplementary_data=input_data.supplementary_data)

    return converted



import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error

from examples.advanced.time_series_forecasting.composing_pipelines import visualise, get_border_line_info
from examples.simple.time_series_forecasting.ts_pipelines import ts_locf_ridge_pipeline
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from fedot.core.utils import fedot_project_root

datasets = {
    'australia': f'{fedot_project_root()}/examples/data/ts/australia.csv',
    'beer': f'{fedot_project_root()}/examples/data/ts/beer.csv',
    'salaries': f'{fedot_project_root()}/examples/data/ts/salaries.csv',
    'stackoverflow': f'{fedot_project_root()}/examples/data/ts/stackoverflow.csv'}


def run_experiment(dataset: str, pipeline: Pipeline, len_forecast=250, tuning=True, visualisalion=False):
    """ Example of ts forecasting using custom pipelines with optional tuning
    :param dataset: name of dataset
    :param pipeline: pipeline to use
    :param len_forecast: forecast length
    :param tuning: is tuning needed
    """
    # show initial pipeline
    pipeline.print_structure()

    time_series = pd.read_csv(datasets[dataset])

    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=len_forecast))
    if dataset not in ['australia']:
        idx = pd.to_datetime(time_series['idx'].values)
    else:
        # non datetime indexes
        idx = time_series['idx'].values
    time_series = time_series['value'].values
    train_input = InputData(idx=idx,
                            features=time_series,
                            target=time_series,
                            task=task,
                            data_type=DataTypesEnum.ts)
    train_data, test_data = train_test_data_setup(train_input)
    test_target = np.ravel(test_data.target)

    pipeline.fit(train_data)

    prediction = pipeline.predict(test_data)
    predict = np.ravel(np.array(prediction.predict))

    plot_info = []
    metrics_info = {}
    plot_info.append({'idx': idx,
                      'series': time_series,
                      'label': 'Actual time series'})

    rmse = mean_squared_error(test_target, predict, squared=False)
    mae = mean_absolute_error(test_target, predict)

    metrics_info['Metrics without tuning'] = {'RMSE': round(rmse, 3),
                                              'MAE': round(mae, 3)}
    plot_info.append({'idx': prediction.idx,
                      'series': predict,
                      'label': 'Forecast without tuning'})
    plot_info.append(get_border_line_info(prediction.idx[0], predict, time_series, 'Border line'))

    if tuning:
        tuner = TunerBuilder(task)\
            .with_tuner(PipelineTuner)\
            .with_metric(RegressionMetricsEnum.MSE)\
            .with_iterations(100) \
            .build(train_data)
        pipeline = tuner.tune(pipeline)
        pipeline.fit(train_data)
        prediction_after = pipeline.predict(test_data)
        predict_after = np.ravel(np.array(prediction_after.predict))

        rmse = mean_squared_error(test_target, predict_after, squared=False)
        mae = mean_absolute_error(test_target, predict_after)

        metrics_info['Metrics after tuning'] = {'RMSE': round(rmse, 3),
                                                'MAE': round(mae, 3)}
        plot_info.append({'idx': prediction_after.idx,
                          'series': predict_after,
                          'label': 'Forecast after tuning'})

    print(metrics_info)
    # plot lines
    if visualisalion:
        visualise(plot_info)
        pipeline.print_structure()


if __name__ == '__main__':
    run_experiment('australia', ts_locf_ridge_pipeline(), len_forecast=50, tuning=True, visualisalion=True)



import datetime
from copy import deepcopy

import numpy as np
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error

from cases.multi_ts_level_forecasting import prepare_data
from examples.simple.time_series_forecasting.ts_pipelines import ts_complex_ridge_smoothing_pipeline
from fedot.core.composer.composer_builder import ComposerBuilder
from fedot.core.composer.gp_composer.specific_operators import parameter_change_mutation
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.mutation import MutationTypesEnum
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.quality_metrics_repository import \
    RegressionMetricsEnum


def calculate_metrics(target, predicted):
    rmse = mean_squared_error(target, predicted, squared=True)
    mae = mean_absolute_error(target, predicted)
    return rmse, mae


def get_available_operations():
    """ Function returns available operations for primary and secondary nodes """
    primary_operations = ['lagged', 'smoothing', 'diff_filter', 'gaussian_filter']
    secondary_operations = ['lagged', 'ridge', 'lasso', 'linear']
    return primary_operations, secondary_operations


def compose_pipeline(pipeline, train_data, task):
    # pipeline structure optimization
    primary_operations, secondary_operations = get_available_operations()
    composer_requirements = PipelineComposerRequirements(
        primary=primary_operations,
        secondary=secondary_operations,
        max_arity=3, max_depth=5,
        num_of_generations=30,
        timeout=datetime.timedelta(minutes=10))
    optimizer_parameters = GPGraphOptimizerParameters(
        pop_size=15,
        mutation_prob=0.8, crossover_prob=0.8,
        mutation_types=[parameter_change_mutation,
                        MutationTypesEnum.single_change,
                        MutationTypesEnum.single_drop,
                        MutationTypesEnum.single_add]
    )
    composer = ComposerBuilder(task=task). \
        with_optimizer_params(optimizer_parameters). \
        with_requirements(composer_requirements). \
        with_metrics(RegressionMetricsEnum.MAE). \
        with_initial_pipelines([pipeline]). \
        build()
    obtained_pipeline = composer.compose_pipeline(data=train_data)
    obtained_pipeline.show()
    return obtained_pipeline


def run_multiple_ts_forecasting(forecast_length, is_multi_ts):
    # separate data on test/train
    train_data, test_data, task = prepare_data(forecast_length, is_multi_ts=is_multi_ts)
    # pipeline initialization
    pipeline = ts_complex_ridge_smoothing_pipeline()
    # pipeline fit and predict
    pipeline.fit(train_data)
    prediction_before = np.ravel(np.array(pipeline.predict(test_data).predict))
    # metrics evaluation
    rmse, mae = calculate_metrics(np.ravel(test_data.target), prediction_before)

    # compose pipeline with initial approximation
    obtained_pipeline = compose_pipeline(pipeline, train_data, task)
    # composed pipeline fit and predict
    obtained_pipeline_copy = deepcopy(obtained_pipeline)
    obtained_pipeline.fit_from_scratch(train_data)
    prediction_after = obtained_pipeline.predict(test_data)
    predict_after = np.ravel(np.array(prediction_after.predict))
    # metrics evaluation
    rmse_composing, mae_composing = calculate_metrics(np.ravel(test_data.target), predict_after)

    # tuning composed pipeline
    tuner = TunerBuilder(task)\
        .with_tuner(PipelineTuner)\
        .with_metric(RegressionMetricsEnum.MAE)\
        .with_iterations(50)\
        .build(train_data)
    obtained_pipeline_copy = tuner.tune(obtained_pipeline_copy)
    obtained_pipeline_copy.print_structure()
    # tuned pipeline fit and predict
    obtained_pipeline_copy.fit_from_scratch(train_data)
    prediction_after_tuning = obtained_pipeline_copy.predict(test_data)
    predict_after_tuning = np.ravel(np.array(prediction_after_tuning.predict))
    # metrics evaluation
    rmse_tuning, mae_tuning = calculate_metrics(np.ravel(test_data.target), predict_after_tuning)

    # visualization of results
    if is_multi_ts:
        history = np.ravel(train_data.target[:, 0])
    else:
        history = np.ravel(train_data.target)
    plt.plot(np.ravel(test_data.idx), np.ravel(test_data.target), label='test')
    plt.plot(np.ravel(train_data.idx), history, label='history')
    plt.plot(np.ravel(test_data.idx), prediction_before, label='prediction')
    plt.plot(np.ravel(test_data.idx), predict_after, label='prediction_after_composing')
    plt.plot(np.ravel(test_data.idx), predict_after_tuning, label='prediction_after_tuning')
    plt.xlabel('Time step')
    plt.ylabel('Sea level')
    plt.legend()
    plt.show()

    print(f'RMSE: {round(rmse, 3)}')
    print(f'MAE: {round(mae, 3)}')
    print(f'RMSE after composing: {round(rmse_composing, 3)}')
    print(f'MAE after composing: {round(mae_composing, 3)}')
    print(f'RMSE after tuning: {round(rmse_tuning, 3)}')
    print(f'MAE after tuning: {round(mae_tuning, 3)}')


if __name__ == '__main__':
    run_multiple_ts_forecasting(forecast_length=60, is_multi_ts=True)



from os import makedirs
from os.path import join, exists

from examples.advanced.sensitivity_analysis.pipelines_access import pipeline_by_task
from fedot.core.data.data import InputData
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.node_sa_approaches import NodeDeletionAnalyze, NodeReplaceOperationAnalyze
from fedot.sensitivity.operations_hp_sensitivity.multi_operations_sensitivity import MultiOperationsHPAnalyze
from fedot.sensitivity.nodes_sensitivity import NodesAnalysis
from fedot.sensitivity.pipeline_sensitivity_facade import PipelineSensitivityAnalysis
from fedot.sensitivity.pipeline_sensitivity import PipelineAnalysis

SA_CLASS_WITH_APPROACHES = {'PipelineSensitivityAnalysis': {'class': PipelineSensitivityAnalysis,
                                                         'approaches': [NodeDeletionAnalyze,
                                                                        NodeReplaceOperationAnalyze,
                                                                        MultiOperationsHPAnalyze]},
                            'NodesAnalysis': {'class': NodesAnalysis,
                                              'approaches': [NodeDeletionAnalyze,
                                                             NodeReplaceOperationAnalyze]},
                            'PipelineAnalysis': {'class': PipelineAnalysis,
                                              'approaches': [MultiOperationsHPAnalyze]},

                            }


def run_case_analysis(train_data: InputData, test_data: InputData,
                      case_name: str, task, metric, sa_class,
                      is_composed=False, result_path=None):
    pipeline = pipeline_by_task(task=task, metric=metric,
                          data=train_data, is_composed=is_composed)

    pipeline.fit(train_data)

    if not result_path:
        result_path = join(default_fedot_data_dir(), 'sensitivity', f'{case_name}')
        if not exists(result_path):
            makedirs(result_path)
    pipeline.show(join(result_path, f'{case_name}'))

    sa_class_with_approaches = SA_CLASS_WITH_APPROACHES.get(sa_class)
    sa_class = sa_class_with_approaches['class']
    approaches = sa_class_with_approaches['approaches']

    pipeline_analysis_result = sa_class(pipeline=pipeline,
                                        train_data=train_data,
                                        test_data=test_data,
                                        approaches=approaches,
                                        path_to_save=result_path).analyze()

    print(f'pipeline analysis result {pipeline_analysis_result}')



import datetime
import os
import random

import numpy as np
import pandas as pd

from fedot.core.adapter import DirectAdapter, register_native
from fedot.core.dag.graph_utils import ordered_subnodes_hierarchy
from fedot.core.dag.convert import graph_structure_as_nx_graph
from fedot.core.dag.verification_rules import has_no_cycle, has_no_self_cycled_nodes
from fedot.core.log import default_log
from fedot.core.optimisers.gp_comp.gp_optimizer import EvoGraphOptimizer
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.crossover import CrossoverTypesEnum
from fedot.core.optimisers.gp_comp.operators.inheritance import GeneticSchemeTypesEnum
from fedot.core.optimisers.gp_comp.operators.regularization import RegularizationTypesEnum
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.graph import OptGraph, OptNode
from fedot.core.optimisers.objective import Objective, ObjectiveEvaluate
from fedot.core.optimisers.optimizer import GraphGenerationParams
from fedot.core.utils import fedot_project_root

random.seed(1)
np.random.seed(1)


class CustomGraphModel(OptGraph):
    def evaluate(self, data: pd.DataFrame):
        nodes = data.columns.to_list()
        _, labels = graph_structure_as_nx_graph(self)
        return len(nodes)


class CustomGraphNode(OptNode):
    def __str__(self):
        return f'Node_{self.content["name"]}'


def custom_metric(graph: CustomGraphModel, data: pd.DataFrame):
    graph.show()
    existing_variables_num = -graph.depth - graph.evaluate(data)

    return existing_variables_num


def _has_no_duplicates(graph):
    _, labels = graph_structure_as_nx_graph(graph)
    if len(labels.values()) != len(set(labels.values())):
        raise ValueError('Custom graph has duplicates')
    return True


@register_native
def custom_mutation(graph: OptGraph, **kwargs):
    num_mut = 10
    try:
        for _ in range(num_mut):
            rid = random.choice(range(graph.length))
            random_node = graph.nodes[rid]
            other_random_node = graph.nodes[random.choice(range(len(graph.nodes)))]
            nodes_not_cycling = (random_node.descriptive_id not in
                                 [n.descriptive_id for n in ordered_subnodes_hierarchy(other_random_node)] and
                                 other_random_node.descriptive_id not in
                                 [n.descriptive_id for n in ordered_subnodes_hierarchy(random_node)])
            if nodes_not_cycling:
                graph.connect_nodes(random_node, other_random_node)
    except Exception as ex:
        default_log(prefix='custom_mutation').warning(f'Incorrect connection: {ex}')
    return graph


def run_custom_example(timeout: datetime.timedelta = None):
    if not timeout:
        timeout = datetime.timedelta(minutes=1)
    data = pd.read_csv(os.path.join(fedot_project_root(), 'examples', 'data', 'custom_encoded.csv'))
    nodes_types = ['V1', 'V2', 'V3',
                   'V4', 'V5', 'V6',
                   'V7', 'V8', 'V9', 'V10']
    rules = [has_no_self_cycled_nodes, has_no_cycle, _has_no_duplicates]

    initial = [CustomGraphModel(nodes=[CustomGraphNode(nodes_from=None,
                                                       content={'name': node_type}) for node_type in nodes_types])]

    requirements = PipelineComposerRequirements(
        primary=nodes_types,
        secondary=nodes_types,
        max_arity=10,
        max_depth=10,
        num_of_generations=5,
        timeout=timeout
    )

    optimiser_parameters = GPGraphOptimizerParameters(
        pop_size=5,
        crossover_prob=0.8, mutation_prob=0.9,
        genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
        mutation_types=[custom_mutation],
        crossover_types=[CrossoverTypesEnum.none],
        regularization_type=RegularizationTypesEnum.none)

    graph_generation_params = GraphGenerationParams(
        adapter=DirectAdapter(base_graph_class=CustomGraphModel, base_node_class=CustomGraphNode),
        rules_for_constraint=rules)

    objective = Objective({'custom': custom_metric})
    optimiser = EvoGraphOptimizer(
        graph_generation_params=graph_generation_params,
        objective=objective,
        graph_optimizer_params=optimiser_parameters,
        requirements=requirements, initial_graphs=initial)

    objective_eval = ObjectiveEvaluate(objective, data=data)
    optimized_graphs = optimiser.optimise(objective_eval)
    optimized_network = optimiser.graph_generation_params.adapter.restore(optimized_graphs[0])

    optimized_network.show()


if __name__ == '__main__':
    run_custom_example()



from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Union

import numpy as np

from fedot.core.dag.linked_graph_node import LinkedGraphNode
from fedot.core.data.data import InputData, OutputData
from fedot.core.data.merge.data_merger import DataMerger
from fedot.core.log import default_log
from fedot.core.operations.factory import OperationFactory
from fedot.core.operations.operation import Operation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.optimisers.timer import Timer
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.serializers.serializer import register_serializable
from fedot.core.utils import DEFAULT_PARAMS_STUB


@register_serializable
@dataclass
class NodeMetadata:
    """Dataclass. :class:`Node` metadata

    Args:
        metric: quality score
    """

    metric: Optional[float] = None


class Node(LinkedGraphNode):
    """Base class for node definition in :class:`Pipeline` structure

    Args:
        nodes_from: parent nodes which information comes from
        operation_type: type of the operation defined in operation repository
            the custom prefix can be added after ``/`` (to highlight the specific node)\n
            **The prefix will be ignored at Implementation stage**
    """

    def __init__(self, nodes_from: Optional[List['Node']],
                 operation_type: Optional[Union[str, 'Operation']] = None, **kwargs):
        passed_content = kwargs.get('content')
        if passed_content:
            # Define operation, based on content dictionary
            operation = self._process_content_init(passed_content)
            params = passed_content.get('params', {})
            # The check for "default_params" is needed for backward compatibility.
            if params == DEFAULT_PARAMS_STUB:
                params = {}
            self.metadata = passed_content.get('metadata', NodeMetadata())
        else:
            # There is no content for node
            operation = self._process_direct_init(operation_type)

            # Define operation with default parameters
            params = {}
            self.metadata = NodeMetadata()

        self.fit_time_in_seconds = 0
        self.inference_time_in_seconds = 0

        self._parameters = OperationParameters.from_operation_type(operation.operation_type, **params)
        super().__init__(content={'name': operation,
                                  'params': self._parameters.to_dict()}, nodes_from=nodes_from)
        self.log = default_log(self)
        self._fitted_operation = None
        self.rating = None

    def _process_content_init(self, passed_content: dict) -> Operation:
        """ Updating content in the node """
        if isinstance(passed_content['name'], str):
            # Need to convert name of operation into operation class object
            operation_factory = OperationFactory(operation_name=passed_content['name'])
            operation = operation_factory.get_operation()

            passed_content.update({'name': operation})
        else:
            operation = passed_content['name']
        self.content = passed_content

        return operation

    @staticmethod
    def _process_direct_init(operation_type: Optional[Union[str, Operation]]) -> Operation:
        """Define operation based on the direct ``operation_type`` without defining content in the node

        Args:
            operation_type: node type representation

        Returns:
            Operation: operation class object
        """
        if not operation_type:
            raise ValueError('Operation is not defined in the node')

        if not isinstance(operation_type, str):
            # AtomizedModel
            operation = operation_type
        else:
            # Define appropriate operation or data operation
            operation_factory = OperationFactory(operation_name=operation_type)
            operation = operation_factory.get_operation()

        return operation

    def update_params(self):
        """Updates :attr:`custom_params` with changed parameters"""
        new_params = self.fitted_operation.get_params()
        changed_parameters = new_params.changed_parameters
        updated_parameters = {**self.parameters, **changed_parameters}
        self.parameters = updated_parameters

    @property
    def name(self) -> str:
        """ Returns str name of operation """
        return self.operation.operation_type

    @property
    def operation(self) -> Operation:
        """Returns node operation object

        Returns:
            Operation: operation object
        """
        return self.content['name']

    @operation.setter
    def operation(self, value: Operation):
        """Updates ``operation`` property with the provided ``value``

        Args:
            value: new operation object
        """

        self.content.update({'name': value})

    @property
    def fitted_operation(self) -> Optional[Any]:
        """Returns already fitted operation if exists or ``None`` instead

        Returns:
            node fitted operation or ``None``
        """

        return getattr(self, '_fitted_operation', None)

    @fitted_operation.setter
    def fitted_operation(self, value: Any):
        """Sets node fitted operation with the provided ``value``

        Args:
            value: any model from the ``list`` of acceptable nodes for the chosen task and problem
        """

        if value is None:
            if hasattr(self, '_fitted_operation'):
                del self._fitted_operation
        else:
            self._fitted_operation = value

    def unfit(self):
        """Sets :obj:`fitted_operation` to ``None``

        Todo:
            check how it would be rendered
        """

        self.fitted_operation = None

    def fit(self, input_data: InputData) -> OutputData:
        """Runs training process in the node

        Args:
            input_data: data used for operation training

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """
        if self.fitted_operation is None:
            with Timer() as t:
                self.fitted_operation, operation_predict = self.operation.fit(params=self._parameters,
                                                                              data=input_data)
                self.fit_time_in_seconds = round(t.seconds_from_start, 3)
        else:

            operation_predict = self.operation.predict_for_fit(fitted_operation=self.fitted_operation,
                                                               data=input_data,
                                                               params=self._parameters)

        # Update parameters after operation fitting (they can be corrected)
        not_atomized_operation = 'atomized' not in self.operation.operation_type

        if not_atomized_operation and 'correct_params' in self.operation.metadata.tags:
            self.update_params()
        return operation_predict

    def predict(self, input_data: InputData, output_mode: str = 'default') -> OutputData:
        """Runs prediction process in the node

        Args:
            input_data: data used for prediction
            output_mode: desired output for operations (e.g. ``'labels'``, ``'probs'``, ``'full_probs'``)

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        with Timer() as t:
            operation_predict = self.operation.predict(fitted_operation=self.fitted_operation,
                                                       params=self._parameters,
                                                       data=input_data,
                                                       output_mode=output_mode)
            self.inference_time_in_seconds = round(t.seconds_from_start, 3)
        return operation_predict

    @property
    def parameters(self) -> dict:
        """Returns node custom parameters

        Returns:
            dict: of custom parameters
        """
        return self.content.get('params')

    @parameters.setter
    def parameters(self, params: dict):
        """Sets custom parameters of the node

        Args:
            params: new parameters to be placed instead of existing
        """
        if params:
            # The check for "default_params" is needed for backward compatibility.
            if params == DEFAULT_PARAMS_STUB:
                params = {}
            # take nested composer params if they appeared
            if 'nested_space' in params:
                params = params['nested_space']
            self._parameters = OperationParameters.from_operation_type(self.operation.operation_type, **params)
            self.content['params'] = self._parameters.to_dict()

    def __str__(self) -> str:
        """Returns ``str`` representation of the node

        Returns:
            str: stringified node operation type
        """

        return str(self.operation.operation_type)

    @property
    def tags(self) -> Optional[List[str]]:
        """Returns tags of operation in the node or empty list

        Returns:
            Optional[List[str]]: ``empty list`` if node is of atomized type and ``list of tags`` otherwise
        """

        if 'atomized' in self.operation.operation_type:
            # There are no tags for atomized operation
            return []

        info = OperationTypesRepository(operation_type='all').operation_info_by_id(self.operation.operation_type)
        if info is not None:
            return info.tags


class PrimaryNode(Node):
    """Defines the interface of primary nodes where initial task data is located

    Args:
        operation_type: operation defined in the operation repository
        node_data: ``dict`` with :class:`InputData` for fit and predict stage
        kwargs: optional arguments (i.e. logger); ``nodes_from`` will be deleted if exists
    """

    def __init__(self, operation_type: Optional[Union[str, 'Operation']] = None, node_data: Optional[dict] = None,
                 **kwargs):
        if 'nodes_from' in kwargs:
            del kwargs['nodes_from']

        super().__init__(nodes_from=None, operation_type=operation_type, **kwargs)

        if node_data is None:
            self._node_data = {}
            self.direct_set = False
        else:
            self._node_data = node_data
            # Was the data passed directly to the node or not
            self.direct_set = True

    def fit(self, input_data: InputData, **kwargs) -> OutputData:
        """Fits the operation located in the primary node

        Args:
            input_data: data used for operation training

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        self.log.debug(f'Trying to fit primary node with operation: {self.operation}')

        if self.direct_set:
            input_data = self.node_data
        else:
            self.node_data = input_data
        return super().fit(input_data)

    def unfit(self):
        """Sets ``node_data`` (if exists) and ``fitted_operation`` to ``None``
        """

        self.fitted_operation = None
        if hasattr(self, 'node_data'):
            self.node_data = None

    def predict(self, input_data: InputData, output_mode: str = 'default') -> OutputData:
        """Predicts using the operation located in the primary node

        Args:
            input_data: data used for prediction
            output_mode: desired output for operations (e.g. ``'labels'``, ``'probs'``, ``'full_probs'``)

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        self.log.debug(f'Predict in primary node by operation: {self.operation}')

        if self.direct_set:
            input_data = self.node_data
        else:
            self.node_data = input_data
        return super().predict(input_data, output_mode)

    def get_data_from_node(self) -> dict:
        """Returns data if it was set to the nodes directly

        Returns:
            dict: ``dict`` with :class:`InputData` for fit and predict stage
        """

        return self.node_data

    @property
    def node_data(self) -> dict:
        """Returns directly set :attr:`node_data`

        Returns:
            dict: ``dict`` with :class:`InputData` for fit and predict stage
        """

        return getattr(self, '_node_data', {})

    @node_data.setter
    def node_data(self, value: dict):
        """Sets :attr:`node_data`

        Args:
            value: ``dict`` with :class:`InputData` for fit and predict stage
        """

        if value is None:
            if hasattr(self, '_node_data'):
                del self._node_data
        else:
            self._node_data = value


class SecondaryNode(Node):
    """The class defines the interface of ``Secondary`` nodes modifying tha data flow in the :class:`Pipeline`

    Args:
        operation_type: operation defined in the operation repository
        nodes_from: parent nodes where data comes from
        kwargs: optional arguments (i.e. logger)
    """

    def __init__(self, operation_type: Optional[Union[str, 'Operation']] = None,
                 nodes_from: Optional[List['Node']] = None, **kwargs):
        super().__init__(nodes_from=nodes_from, operation_type=operation_type, **kwargs)

    def fit(self, input_data: InputData, **kwargs) -> OutputData:
        """Fits the operation located in the secondary node

        Args:
            input_data: data used for operation training

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        self.log.debug(f'Trying to fit secondary node with operation: {self.operation}')

        secondary_input = self._input_from_parents(input_data=input_data, parent_operation='fit')
        return super().fit(input_data=secondary_input)

    def predict(self, input_data: InputData, output_mode: str = 'default') -> OutputData:
        """Predicts using the operation located in the secondary node

        Args:
            input_data: data used for prediction
            output_mode: desired output for operations (e.g. ``'labels'``, ``'probs'``, ``'full_probs'``)

        Returns:
            OutputData: values predicted on the provided ``input_data``
        """

        self.log.debug(f'Obtain prediction in secondary node with operation: {self.operation}')

        secondary_input = self._input_from_parents(input_data=input_data,
                                                   parent_operation='predict')

        return super().predict(input_data=secondary_input, output_mode=output_mode)

    def _input_from_parents(self, input_data: InputData, parent_operation: str) -> InputData:
        """Processes all the parent nodes via the current operation using ``input_data``

        Args:
            input_data: input data from pipeline abstraction (source input data)
            parent_operation: name of parent operation (``'fit'`` or ``'predict'``)

        Returns:
            InputData: predictions from the secondary nodes
        """

        if len(self.nodes_from) == 0:
            raise ValueError('No parent nodes found')

        self.log.debug(f'Fit all parent nodes in secondary node with operation: {self.operation}')

        parent_nodes = self._nodes_from_with_fixed_order()

        parent_results, _ = _combine_parents(parent_nodes, input_data,
                                             parent_operation)
        secondary_input = DataMerger.get(parent_results).merge()
        # Update info about visited nodes
        parent_operations = [node.operation.operation_type for node in parent_nodes]
        secondary_input.supplementary_data.previous_operations = parent_operations
        return secondary_input

    def _nodes_from_with_fixed_order(self):
        """Sorts :attr:`nodes_from` (if exists) by the nodes unique id

        Returns:
            sorted :attr:`nodes_from` by :obj:`GraphNode.descriptive_id` or ``None``
        """
        return sorted(self.nodes_from, key=lambda node: node.descriptive_id)


def _combine_parents(parent_nodes: List[Node],
                     input_data: Optional[InputData], parent_operation: str) -> Tuple[List[OutputData], np.array]:
    """Сombines predictions from the ``parent_nodes``

    Args:
        parent_nodes: list of parent nodes, from which predictions will be combined
        input_data: input data from pipeline abstraction (source input data)
        parent_operation: name of parent operation (``'fit'`` or ``'predict'``)

    Returns:
        Tuple[List[OutputData], np.array]: :obj:`output data list from parent nodes`,
        :obj:`target for final pipeline prediction`
    """

    if input_data is not None:
        # InputData was set to pipeline
        target = input_data.target
    parent_results = []
    for parent in parent_nodes:

        if parent_operation == 'predict':
            prediction = parent.predict(input_data=input_data)
            parent_results.append(prediction)
        elif parent_operation == 'fit':
            prediction = parent.fit(input_data=input_data)
            parent_results.append(prediction)
        else:
            raise NotImplementedError()

        if input_data is None:
            # InputData was set to primary nodes
            target = prediction.target

    return parent_results, target



from dataclasses import dataclass
from typing import Sequence, Union, Any

from fedot.core.optimisers.optimizer import GraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.crossover import CrossoverTypesEnum
from fedot.core.optimisers.gp_comp.operators.elitism import ElitismTypesEnum
from fedot.core.optimisers.gp_comp.operators.inheritance import GeneticSchemeTypesEnum
from fedot.core.optimisers.gp_comp.operators.mutation import MutationTypesEnum, MutationStrengthEnum
from fedot.core.optimisers.gp_comp.operators.regularization import RegularizationTypesEnum
from fedot.core.optimisers.gp_comp.operators.selection import SelectionTypesEnum


@dataclass
class GPGraphOptimizerParameters(GraphOptimizerParameters):
    """
    Defines parameters of evolutionary operators and the algorithm of genetic optimizer.

    :param crossover_prob: crossover probability (the chance that two chromosomes exchange some of their parts)
    :param mutation_prob: mutation probability
    :param static_mutation_prob: probability of applying same mutation to graph in a cycle of mutations
    :param max_num_of_operator_attempts: max number of unsuccessful operator (mutation/crossover)
    attempts before continuing
    :param mutation_strength: strength of mutation in tree (using in certain mutation types)
    :param min_pop_size_with_elitism: minimal population size with which elitism is applicable

    :param selection_types: Sequence of selection operators types
    :param crossover_types: Sequence of crossover operators types
    :param mutation_types: Sequence of mutation operators types
    :param elitism_type: type of elitism operator evolution
    :param regularization_type: type of regularization operator
    :param genetic_scheme_type: type of genetic evolutionary scheme
    """

    crossover_prob: float = 0.8
    mutation_prob: float = 0.8
    static_mutation_prob: float = 0.7
    max_num_of_operator_attempts: int = 100
    mutation_strength: MutationStrengthEnum = MutationStrengthEnum.mean
    min_pop_size_with_elitism: int = 5

    selection_types: Sequence[SelectionTypesEnum] = \
        (SelectionTypesEnum.tournament,)
    crossover_types: Sequence[Union[CrossoverTypesEnum, Any]] = \
        (CrossoverTypesEnum.subtree,
         CrossoverTypesEnum.one_point)
    mutation_types: Sequence[Union[MutationTypesEnum, Any]] = \
        (MutationTypesEnum.simple,
         MutationTypesEnum.reduce,
         MutationTypesEnum.growth,
         MutationTypesEnum.local_growth)
    elitism_type: ElitismTypesEnum = ElitismTypesEnum.keep_n_best
    regularization_type: RegularizationTypesEnum = RegularizationTypesEnum.none
    genetic_scheme_type: GeneticSchemeTypesEnum = GeneticSchemeTypesEnum.generational

    def __post_init__(self):
        if self.multi_objective:
            self.selection_types = (SelectionTypesEnum.spea2,)
            # TODO add possibility of using regularization in MO alg
            self.regularization_type = RegularizationTypesEnum.none



import os
import json


class DefaultOperationParamsRepository:
    def __init__(self, repository_name: str = 'default_operation_params.json'):
        repo_folder_path = str(os.path.dirname(__file__))
        file = os.path.join('data', repository_name)
        self._repo_path = os.path.join(repo_folder_path, file)
        self._repo = self._initialise_repo()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._repo_path = None

    def _initialise_repo(self) -> dict:
        with open(self._repo_path) as repository_json_file:
            repository_json = json.load(repository_json_file)

        return repository_json

    def get_default_params_for_operation(self, model_name: str) -> dict:
        model_name = model_name.split('/')[0]
        if model_name in self._repo:
            return self._repo[model_name]
        return {}



import os
import platform
import tempfile

from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd

from sklearn.model_selection import train_test_split

DEFAULT_PARAMS_STUB = 'default_params'


def copy_doc(source_func: Callable) -> Callable:
    """
    Copies a docstring from the provided ``source_func`` to the wrapped function

    :param source_func: function to copy the docstring from

    :return: wrapped function with the same docstring as in the given ``source_func``
    """
    def wrapper(func: Callable) -> Callable:
        func.__doc__ = source_func.__doc__
        return func
    return wrapper


def fedot_project_root() -> Path:
    """Returns FEDOT project root folder."""
    return Path(__file__).parent.parent.parent


def default_fedot_data_dir() -> str:
    """ Returns the folder where all the output data
    is recorded to. Default: home/Fedot
    """
    temp_folder = Path("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())
    default_data_path = os.path.join(temp_folder, 'FEDOT')

    if 'FEDOT' not in os.listdir(temp_folder):
        os.mkdir(default_data_path)

    return default_data_path


def labels_to_dummy_probs(prediction: np.array):
    """ Returns converted predictions using one-hot probability encoding """
    df = pd.Series(prediction)
    pred_probas = pd.get_dummies(df).values
    return pred_probas


def probs_to_labels(prediction: np.array):
    """ Converts predicted probabilities into labels """
    list_with_labels = []
    for list_with_probs in prediction:
        list_with_labels.append(list_with_probs.argmax())

    return np.asarray(list_with_labels).reshape((-1, 1))


def split_data(df: pd.DataFrame, t_size: float = 0.2):
    """ Split pandas DataFrame into train and test parts """
    train, test = train_test_split(df.iloc[:, :], test_size=t_size, random_state=42)
    return train, test


def save_file_to_csv(df: pd.DataFrame, path_to_save: str):
    return df.to_csv(path_to_save, sep=',')


def get_split_data_paths(directory_names: list):
    train_file_path = os.path.join(directory_names[0], directory_names[1], directory_names[2], 'train.csv')
    full_train_file_path = os.path.join(str(fedot_project_root()), train_file_path)
    test_file_path = os.path.join(directory_names[0], directory_names[1], directory_names[2], 'test.csv')
    full_test_file_path = os.path.join(str(fedot_project_root()), test_file_path)
    return full_train_file_path, full_test_file_path


def ensure_directory_exists(dir_names: list):
    main_dir = os.path.join(str(fedot_project_root()), dir_names[0], dir_names[1])
    dataset_dir = os.path.join(str(fedot_project_root()), dir_names[0], dir_names[1], dir_names[2])
    if not os.path.exists(main_dir):
        os.mkdir(main_dir)
    if not os.path.exists(dataset_dir):
        os.mkdir(dataset_dir)


def make_pipeline_generator(pipeline):
    visited_nodes = []

    for node in pipeline.nodes:
        if node not in visited_nodes:
            visited_nodes.append(node)
            yield node



import os
import pickle
from pathlib import Path
from typing import TYPE_CHECKING
from uuid import uuid4

if TYPE_CHECKING:
    from fedot.core.data.data import InputData
    from fedot.core.pipelines.pipeline import Pipeline

from fedot.core.utils import default_fedot_data_dir


def is_test_session():
    return 'PYTEST_CURRENT_TEST' in os.environ


def is_recording_mode():
    return 'FEDOT_RECORDING_MODE' in os.environ


def save_debug_info_for_pipeline(pipeline: 'Pipeline', train_data: 'InputData', test_data: 'InputData',
                                 exception: Exception, stack_trace: str):
    try:
        tmp_folder = Path(default_fedot_data_dir(), 'debug_info')
        if not tmp_folder.exists():
            os.mkdir(tmp_folder)

        pipeline_id = str(uuid4())
        base_path = Path(tmp_folder, pipeline_id)
        pipeline.save(f'{base_path}_pipeline', is_datetime_in_path=False)

        with open(f'{base_path}_train_data.pkl', 'wb') as file:
            pickle.dump(train_data, file)
        with open(f'{base_path}_test_data.pkl', 'wb') as file:
            pickle.dump(test_data, file)
        with open(f'{base_path}_exception.txt', 'w') as file:
            print(exception, file=file)
            print(stack_trace, file=file)
    except Exception as ex:
        print(ex)


def reproduce_fitting_error(pipeline_id: str, base_path=None):
    from fedot.core.pipelines.pipeline import Pipeline

    if not base_path:
        base_path = Path(default_fedot_data_dir(), 'debug_info')

    pipeline = Pipeline()
    pipeline.load(str(Path(base_path, f'{pipeline_id}_pipeline', f'{pipeline_id}_pipeline.json')))
    with open(Path(base_path, f'{pipeline_id}_train_data.pkl'), 'rb') as file:
        train_data = pickle.load(file)
    with open(Path(base_path, f'{pipeline_id}_test_data.pkl'), 'rb') as file:
        test_data = pickle.load(file)
    pipeline.unfit()
    pipeline.fit(train_data)
    pipeline.predict(test_data)



from copy import copy

import numpy as np
from sklearn.preprocessing import StandardScaler

from fedot.core.data.data import InputData
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import (
    prepare_target,
    ts_to_table, transform_features_and_target_into_lagged
)
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.pipelines.ts_wrappers import _update_input, exception_if_not_ts_task
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.utilities.requirements_notificator import warn_requirement


class TorchMock:
    Module = list


try:
    import torch
    import torch.nn as nn

    from torch.utils.data import DataLoader, TensorDataset
except ModuleNotFoundError:
    warn_requirement('torch')
    torch = object()
    nn = TorchMock


class CLSTMImplementation(ModelImplementation):
    def __init__(self, params: OperationParameters):
        super().__init__(params)
        self.device = self._get_device()
        self.model = LSTMNetwork(
            hidden_size=int(params.get("hidden_size")),
            cnn1_kernel_size=int(params.get("cnn1_kernel_size")),
            cnn1_output_size=int(params.get("cnn1_output_size")),
            cnn2_kernel_size=int(params.get("cnn2_kernel_size")),
            cnn2_output_size=int(params.get("cnn2_output_size"))
        )

        self.optim_dict = {
            'adam': torch.optim.Adam(self.model.parameters(), lr=self.learning_rate),
            'sgd': torch.optim.SGD(self.model.parameters(), lr=self.learning_rate)
        }

        self.loss_dict = {
            'mae': nn.L1Loss,
            'mse': nn.MSELoss
        }

        self.scaler = StandardScaler()
        self.optimizer = self.optim_dict[params.get("optimizer")]
        self.criterion = self.loss_dict[params.get("loss")]()

    @property
    def learning_rate(self) -> float:
        return self.params.get("learning_rate")

    @property
    def window_size(self) -> int:
        return self.params.get("window_size")

    def fit(self, train_data: InputData):
        """ Class fit ar model on data.

        Implementation uses the idea of teacher forcing. That means model learns
        to predict data when horizon != 1. It uses real values or previous model output
        to predict next value. self.teacher_forcing param is used to control probability
        of using real y values.

        :param train_data: data with features, target and ids to process
        """

        self.model = self.model.to(self.device)
        data_loader, forecast_length = self._create_dataloader(train_data)

        self.model.train()
        for epoch in range(self.params.get("num_epochs")):
            for x, y in data_loader:
                self.optimizer.zero_grad()
                x = x.to(self.device)
                y = y.to(self.device)
                final_output = self._apply_teacher_forcing(x, y, forecast_length)
                loss = self.criterion(final_output, y)
                loss.backward()
                self.optimizer.step()
        return self.model

    def _apply_teacher_forcing(self, x, y, forecast_length):
        final_output = None
        for i in range(forecast_length):
            self.model.init_hidden(x.shape[0], self.device)
            output = self.model(x.unsqueeze(1)).squeeze(0)
            if np.random.random_sample() > int(self.params.get("teacher_forcing")):
                x = torch.hstack((x[:, 1:], output))
            else:
                x = torch.hstack((x, y[:, i].unsqueeze(1)))

            if final_output is not None:
                final_output = torch.hstack((final_output, output))
            else:
                final_output = output
        return final_output

    def predict(self, input_data: InputData):
        """ Method for time series prediction on forecast length

        :param input_data: data with features, target and ids to process
        :return output_data: output data with smoothed time series
        """
        self.model.eval()
        input_data = copy(input_data)
        forecast_length = input_data.task.task_params.forecast_length

        input_data.features = input_data.features[-self.window_size:].reshape(1, -1)
        input_data.idx = input_data.idx[-forecast_length:]

        predict = self._out_of_sample_ts_forecast(input_data)

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData):
        self.model.eval()
        input_data = copy(input_data)
        forecast_length = input_data.task.task_params.forecast_length
        new_idx, lagged_table = ts_to_table(idx=input_data.idx,
                                            time_series=input_data.features,
                                            window_size=self.window_size,
                                            is_lag=True)

        final_idx, features_columns, final_target = prepare_target(all_idx=input_data.idx,
                                                                   idx=new_idx,
                                                                   features_columns=lagged_table,
                                                                   target=input_data.target,
                                                                   forecast_length=forecast_length)
        input_data.idx = final_idx
        input_data.features = features_columns
        input_data.target = final_target
        predict = self._out_of_sample_ts_forecast(input_data)

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def _predict(self, input_data: InputData):
        features_scaled = self._transform_scaler_features(input_data)
        x = torch.Tensor(features_scaled).to(self.device)
        self.model.init_hidden(x.shape[0], self.device)
        predict = self.model(x.unsqueeze(1)).squeeze(0).cpu().detach().numpy()
        return self._inverse_transform_scaler(predict)

    def _out_of_sample_ts_forecast(self, input_data: InputData) -> np.array:
        """ Method for out_of_sample CLSTM forecasting (use previous outputs as next inputs)

        :param input_data: data with features, target and ids to process
        :return np.array: np.array with predicted values to process it into output_data
        """

        input_data_new = copy(input_data)
        # Prepare data for time series forecasting
        task = input_data_new.task
        exception_if_not_ts_task(task)

        pre_history_ts = np.array(input_data_new.features)

        number_of_iterations = task.task_params.forecast_length

        final_forecast = None

        for _ in range(0, number_of_iterations):
            with torch.no_grad():
                iter_predict = self._predict(input_data_new)
            if final_forecast is not None:
                final_forecast = np.hstack((final_forecast, iter_predict))
            else:
                final_forecast = iter_predict

            # Add prediction to the historical data - update it
            pre_history_ts = np.hstack((pre_history_ts[:, 1:], iter_predict))
            # Prepare InputData for next iteration
            input_data_new = _update_input(pre_history_ts, number_of_iterations, task)

        return final_forecast

    def _fit_transform_scaler(self, data: InputData):
        f_scaled = self.scaler.fit_transform(data.features.reshape(-1, 1)).reshape(-1)
        t_scaled = self.scaler.transform(data.target.reshape(-1, 1)).reshape(-1)
        return f_scaled, t_scaled

    def _inverse_transform_scaler(self, data: np.ndarray):
        start_shape = data.shape
        return self.scaler.inverse_transform(data.reshape(-1, 1)).reshape(start_shape)

    def _transform_scaler_features(self, data: InputData):
        start_shape = data.features.shape
        return self.scaler.transform(data.features.reshape(-1, 1)).reshape(start_shape)

    def _transform_scaler_target(self, data: InputData):
        start_shape = data.features.shape
        return self.scaler.transform(data.target.reshape(-1, 1)).reshape(start_shape)

    @staticmethod
    def _get_device():
        if torch.cuda.is_available():
            device = 'cuda:0'
        else:
            device = 'cpu'
        return device

    def _create_dataloader(self, input_data: InputData):
        """ Method for creating torch.utils.data.DataLoader object from input_data

        Generate lag tables and process it into DataLoader

        :param input_data: data with features, target and ids to process
        :return torch.utils.data.DataLoader: DataLoader with train data
        """
        forecast_length = input_data.task.task_params.forecast_length
        features_scaled, target_scaled = self._fit_transform_scaler(input_data)
        input_data.features = features_scaled
        input_data.target = target_scaled
        final_idx, features_columns, final_target = transform_features_and_target_into_lagged(input_data,
                                                                                              forecast_length,
                                                                                              self.window_size)
        x = torch.from_numpy(features_columns.copy()).float()
        y = torch.from_numpy(final_target.copy()).float()
        return DataLoader(TensorDataset(x, y), batch_size=self.params.get("batch_size")), forecast_length


class LSTMNetwork(nn.Module):
    def __init__(self,
                 hidden_size=200,
                 cnn1_kernel_size=5,
                 cnn1_output_size=16,
                 cnn2_kernel_size=3,
                 cnn2_output_size=32,
                 ):
        super().__init__()

        self.hidden_size = hidden_size
        self.conv_block1 = nn.Sequential(
            nn.Conv1d(in_channels=1, out_channels=cnn1_output_size, kernel_size=cnn1_kernel_size),
            nn.ReLU()
        )
        self.conv_block2 = nn.Sequential(
            nn.Conv1d(in_channels=cnn1_output_size, out_channels=cnn2_output_size, kernel_size=cnn2_kernel_size),
            nn.ReLU()
        )
        self.lstm = nn.LSTM(cnn2_output_size, self.hidden_size, dropout=0.1)
        self.hidden_cell = None
        self.linear = nn.Linear(self.hidden_size * 2, 1)

    def init_hidden(self, batch_size, device):
        self.hidden_cell = (torch.zeros(1, batch_size, self.hidden_size).to(device),
                            torch.zeros(1, batch_size, self.hidden_size).to(device))

    def forward(self, x):
        if self.hidden_cell is None:
            raise Exception
        x = self.conv_block1(x)
        x = self.conv_block2(x)
        x = x.permute(2, 0, 1)
        out, self.hidden_cell = self.lstm(x, self.hidden_cell)
        hidden_cat = torch.cat([self.hidden_cell[0], self.hidden_cell[1]], dim=2)
        predictions = self.linear(hidden_cat)

        return predictions



import warnings
from typing import Optional

import numpy as np

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy
from fedot.core.operations.evaluation.operation_implementations.data_operations.text_pretrained \
    import PretrainedEmbeddingsImplementation
from fedot.core.operations.evaluation.operation_implementations.data_operations.text_preprocessing import (
    TextCleanImplementation
)
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer

from fedot.core.operations.operation_parameters import OperationParameters

warnings.filterwarnings("ignore", category=UserWarning)


class SkLearnTextVectorizeStrategy(EvaluationStrategy):
    __operations_by_types = {
        'tfidf': TfidfVectorizer,
        'cntvect': CountVectorizer,
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.vectorizer = self._convert_to_operation(operation_type)
        self.vectorizer = self.vectorizer(**self.params_for_fit.to_dict())

    def fit(self, train_data: InputData):

        features_list = self._convert_to_one_dim(train_data.features)

        self.vectorizer.fit(features_list)

        return self.vectorizer

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:

        features_list = self._convert_to_one_dim(predict_data.features)
        predicted = trained_operation.transform(features_list).toarray()

        # Convert prediction to output (if it is required)
        converted = self._convert_to_output(predicted, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain TextVectorize strategy for {operation_type}')

    @staticmethod
    def _convert_to_one_dim(array_with_text):
        """ Method converts array with text into one-dimensional list

        :param array_with_text: numpy array or list with text data
        :return features_list: one-dimensional list with text
        """
        features = np.ravel(np.array(array_with_text, dtype=str))
        features_list = list(features)
        return features_list

    @property
    def implementation_info(self) -> str:
        return str(self._convert_to_operation(self.operation_type))


class FedotTextPreprocessingStrategy(EvaluationStrategy):
    __operations_by_types = {
        'text_clean': TextCleanImplementation}

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.text_processor = self._convert_to_operation(operation_type)
        self.text_processor = self.text_processor(self.params_for_fit)

    def fit(self, train_data: InputData):
        """
        This method is used for operation training with the data provided

        :param InputData train_data: data used for operation training
        :return: trained model
        """

        self.text_processor.fit(train_data)
        return self.text_processor

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during predict stage.

        :param trained_operation: trained operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """

        prediction = trained_operation.transform(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        This method used for prediction of the target data during fit stage.

        :param trained_operation: trained operation object
        :param predict_data: data to predict
        :return OutputData: passed data with new predicted target
        """
        prediction = trained_operation.transform_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain custom text preprocessing strategy for {operation_type}')


class GensimTextVectorizeStrategy(EvaluationStrategy):
    __operations_by_types = {
        'word2vec_pretrained': PretrainedEmbeddingsImplementation
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.vectorizer = self._convert_to_operation(operation_type)
        self.vectorizer = self.vectorizer(self.params_for_fit)

    def fit(self, train_data: InputData):
        """ Class doesn't support fit operation

        :param train_data: data with features, target and ids to process
        """

        self.vectorizer.fit(train_data)
        return self.vectorizer

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:

        prediction = trained_operation.transform(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:

        prediction = trained_operation.transform_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain text vectorization strategy for {operation_type}')

    @staticmethod
    def _convert_to_one_dim(array_with_text):
        """ Method converts array with text into one-dimensional list

        :param array_with_text: numpy array or list with text data
        :return features_list: one-dimensional list with text
        """
        features = np.ravel(np.array(array_with_text, dtype=str))
        features_list = list(features)
        return features_list

    @property
    def implementation_info(self) -> str:
        return str(self._convert_to_operation(self.operation_type))



from copy import deepcopy
from typing import List, Union

import numpy as np
from scipy import interpolate

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams


def series_has_gaps_check(gapfilling_method):
    """ Check is time series has gaps or not. Return source array, if not """

    def wrapper(self, input_data, *args, **kwargs):
        input_data = replace_nan_with_label(input_data, label=self.gap_value)
        gap_ids = np.ravel(np.argwhere(input_data == self.gap_value))
        if len(gap_ids) == 0:
            self.log.info(f'Array does not contain values marked as gaps {self.gap_value}')
            return input_data
        else:
            self.log.debug(f'Array contain values marked as gaps {self.gap_value}. Start gap-filling')
            filled_array = gapfilling_method(self, input_data, *args, **kwargs)
            return filled_array

    return wrapper


class SimpleGapFiller:
    """
    Base class used for filling in the gaps in time series with simple methods.
    Methods from the SimpleGapFiller class can be used for comparison with more
    complex models in class ModelGapFiller

    Args:
        gap_value: value, which identify gap elements in array
    """

    def __init__(self, gap_value: float = -100.0):
        self.gap_value = gap_value
        self.log = default_log(self)

    @series_has_gaps_check
    def linear_interpolation(self, input_data: np.array):
        """
        Method allows to restore missing values in an array
        using linear interpolation

        Args:
            input_data: array with gaps

        Returns:
            array without gaps
        """

        output_data = np.array(input_data)
        output_data = replace_nan_with_label(output_data, label=self.gap_value)

        # Process first and last elements in time series
        output_data = self._fill_first_and_last_gaps(input_data, output_data)

        # The indices of the known elements
        non_nan = np.ravel(np.argwhere(output_data != self.gap_value))
        # All known elements in the array
        masked_array = output_data[non_nan]
        f_interploate = interpolate.interp1d(non_nan, masked_array)
        x = np.arange(0, len(output_data))
        output_data = f_interploate(x)
        return output_data

    @series_has_gaps_check
    def local_poly_approximation(self, input_data, degree: int = 2,
                                 n_neighbors: int = 5):
        """Method allows to restore missing values in an array
        using Savitzky-Golay filter

        Args:
            input_data: array with gaps
            degree: degree of a polynomial function
            n_neighbors: number of neighboring known elements of the time
            series that the approximation is based on

        Returns:
            array without gaps
        """

        output_data = np.array(input_data)
        output_data = replace_nan_with_label(output_data, label=self.gap_value)

        i_gaps = np.ravel(np.argwhere(output_data == self.gap_value))

        # Iterately fill in the gaps in the time series
        for gap_index in i_gaps:
            # Indexes of known elements (updated at each iteration)
            i_known = np.argwhere(output_data != self.gap_value)
            i_known = np.ravel(i_known)

            # Based on the indexes we calculate how far from the gap
            # the known values are located
            id_distances = np.abs(i_known - gap_index)

            # Now we know the indices of the smallest values in the array,
            # so sort indexes
            sorted_idx = np.argsort(id_distances)
            nearest_values = []
            nearest_indices = []
            for i in sorted_idx[:n_neighbors]:
                time_index = i_known[i]
                nearest_values.append(output_data[time_index])
                nearest_indices.append(time_index)
            nearest_values = np.array(nearest_values)
            nearest_indices = np.array(nearest_indices)

            local_coefs = np.polyfit(nearest_indices, nearest_values, degree)
            est_value = np.polyval(local_coefs, gap_index)
            output_data[gap_index] = est_value

        return output_data

    @series_has_gaps_check
    def batch_poly_approximation(self, input_data, degree: int = 3,
                                 n_neighbors: int = 10):
        """Method allows to restore missing values in an array using
        batch polynomial approximations.
        Approximation is applied not for individual omissions, but for
        intervals of omitted values

        Args:
            input_data: array with gaps
            degree: degree of a polynomial function
            n_neighbors: the number of neighboring known elements of
            time series that the approximation is based on

        Returns:
            array without gaps
        """

        output_data = np.array(input_data)
        output_data = replace_nan_with_label(output_data, label=self.gap_value)

        # Gap indices
        gap_list = np.ravel(np.argwhere(output_data == self.gap_value))
        new_gap_list = self._parse_gap_ids(gap_list)

        # Iterately fill in the gaps in the time series
        for gap in new_gap_list:
            # Find the center point of the gap
            center_index = int((gap[0] + gap[-1]) / 2)

            # Indexes of known elements (updated at each iteration)
            i_known = np.argwhere(output_data != self.gap_value)
            i_known = np.ravel(i_known)

            # Based on the indexes we calculate how far from the gap
            # the known values are located
            id_distances = np.abs(i_known - center_index)

            # Now we know the indices of the smallest values in the array,
            # so sort indexes
            sorted_idx = np.argsort(id_distances)

            # Nearest known values to the gap
            nearest_values = []
            # And their indexes
            nearest_indices = []
            for i in sorted_idx[:n_neighbors]:
                # Getting the index value for the series - output_data
                time_index = i_known[i]
                # Using this index, we get the value of each of the "neighbors"
                nearest_values.append(output_data[time_index])
                nearest_indices.append(time_index)
            nearest_values = np.array(nearest_values)
            nearest_indices = np.array(nearest_indices)

            # Local approximation by an n-th degree polynomial
            local_coefs = np.polyfit(nearest_indices, nearest_values, degree)

            # Estimate our interval according to the selected coefficients
            est_value = np.polyval(local_coefs, gap)
            output_data[gap] = est_value

        return output_data

    def _parse_gap_ids(self, gap_list: Union[List, np.ndarray]) -> list:
        """Method allows parsing source array with gaps indexes

        Args:
            gap_list: array with indexes of gaps in array

        Returns:
            a list with separated gaps in continuous intervals
        """

        new_gap_list = []
        local_gaps = []
        for index, gap in enumerate(gap_list):
            if index == 0:
                local_gaps.append(gap)
            else:
                prev_gap = gap_list[index - 1]
                if gap - prev_gap > 1:
                    # There is a "gap" between gaps
                    new_gap_list.append(local_gaps)

                    local_gaps = []
                    local_gaps.append(gap)
                else:
                    local_gaps.append(gap)
        new_gap_list.append(local_gaps)

        return new_gap_list

    def _fill_first_and_last_gaps(self, input_data: np.array, output_data: np.array):
        """ Eliminate gaps, which place first or last index in time series """
        non_nan_ids = np.ravel(np.argwhere(output_data != self.gap_value))
        non_nan = output_data[non_nan_ids]
        if np.isclose(input_data[0], self.gap_value):
            # First element is a gap - replace with first known value
            self.log.info(f'First element in the array were replaced by first known value')
            output_data[0] = non_nan[0]
        if np.isclose(input_data[-1], self.gap_value):
            # Last element is a gap - last known value
            self.log.info(f'Last element in the array were replaced by last known value')
            output_data[-1] = non_nan[-1]

        return output_data


class ModelGapFiller(SimpleGapFiller):
    """
    Class used for filling in the gaps in time series

    Args:
        gap_value: value, which mask gap elements in array
        pipeline: TsForecastingPipeline object for filling in the gaps
    """

    def __init__(self, gap_value, pipeline):
        super().__init__(gap_value)
        self.pipeline = pipeline

        # At least 6 elements needed to train pipeline with lagged transformation
        self.min_train_ts_length = 6

    @series_has_gaps_check
    def forward_inverse_filling(self, input_data):
        """Method fills in the gaps in the input array using forward and inverse
        directions of predictions

        Args:
            input_data: data with gaps to filling in the gaps in it

        Returns:
            array without gaps
        """
        output_data = np.array(input_data)
        output_data = replace_nan_with_label(output_data, label=self.gap_value)
        # Gap indices
        gap_list = np.ravel(np.argwhere(output_data == self.gap_value))
        new_gap_list = self._parse_gap_ids(gap_list)

        # Iteratively fill in the gaps in the time series
        for batch_index in range(len(new_gap_list)):

            preds = []
            weights = []
            # Two predictions are generated for each gap - forward and backward
            for direction_function in [self._forward, self._inverse]:
                weights_list, predicted_list = direction_function(output_data,
                                                                  batch_index,
                                                                  new_gap_list)
                weights.append(weights_list)
                preds.append(predicted_list)

            preds = np.array(preds)
            weights = np.array(weights)
            result = np.average(preds, axis=0, weights=weights)

            gap = new_gap_list[batch_index]
            # Replace gaps in an array with prediction values
            output_data[gap] = result

        return output_data

    @series_has_gaps_check
    def forward_filling(self, input_data: Union[List, np.ndarray]):
        """ Method fills in the gaps in the input array using graph with only
        forward direction (i.e. time series forecasting)

        Args:
            input_data: data with gaps to filling in the gaps in it

        Returns:
            array without gaps
        """
        output_data = np.array(input_data)
        output_data = replace_nan_with_label(output_data, label=self.gap_value)

        # Gap indices
        gap_list = np.ravel(np.argwhere(output_data == self.gap_value))
        new_gap_list = self._parse_gap_ids(gap_list)

        # Iterately fill in the gaps in the time series
        for gap in new_gap_list:
            # The entire time series is used for training until the gap
            first_gap_element_id = gap[0]
            timeseries_train_part = output_data[:first_gap_element_id]

            # Make forecast in the gap
            predicted = self.__forecast_in_gap(self.pipeline,
                                               timeseries_train_part,
                                               output_data, gap)

            # Replace gaps in an array with prediction values
            output_data[gap] = predicted
        return output_data

    def _forward(self, output_data, batch_index, new_gap_list):
        """The time series method makes a forward forecast based on the part
        of the time series that is located to the left of the gap.

        Args:
            output_data: one-dimensional array of a time series
            batch_index: index of the interval (batch) with a gap
            new_gap_list: array with nested lists of gap indexes

        Returns:
            weights_list: numpy array with prediction weights for averaging
            predicted: numpy array with prediction values in the gap
        """

        gap = new_gap_list[batch_index]
        first_gap_element_id = gap[0]
        timeseries_train_part = output_data[:first_gap_element_id]

        # Adaptive prediction interval length
        len_gap = len(gap)
        predicted = self.__forecast_in_gap(self.pipeline,
                                           timeseries_train_part,
                                           output_data, gap)
        weights_list = np.arange(len_gap, 0, -1)
        return weights_list, predicted

    def _inverse(self, output_data, batch_index, new_gap_list):
        """The time series method makes an inverse forecast based on the part
        of the time series that is located to the right of the gap.

        Args:
            output_data: one-dimensional array of a time series
            batch_index: index of the interval (batch) with a gap
            new_gap_list: array with nested lists of gap indexes

        Returns:
            weights_list: numpy array with prediction weights for averaging
            predicted_values: numpy array with prediction values in the gap
        """

        gap = new_gap_list[batch_index]
        # Adaptive prediction interval length
        len_gap = len(gap)
        weights_list = np.arange(1, (len_gap + 1), 1)

        first_gap_element_id = gap[0]
        latest_gap_element_id = gap[-1]
        if batch_index == len(new_gap_list) - 1:
            # If the interval with a gap is the last one in the array
            timeseries_train_part = output_data[(latest_gap_element_id + 1):]

            is_gap_in_end_time_series = len(timeseries_train_part) == 0
            is_series_size_not_enough = (len(timeseries_train_part) - len_gap) < self.min_train_ts_length
            if is_gap_in_end_time_series:
                # The gap is last element - take last observed value as predicted
                last_known_value = output_data[first_gap_element_id - 1]
                return weights_list, [last_known_value] * len_gap
            elif is_series_size_not_enough:
                # Number of elements in time series after gap is not enough for
                # model training - interpolation is required
                last_known_value_id = first_gap_element_id - 1 if first_gap_element_id > 0 else 0
                extended_part = output_data[last_known_value_id:]
        else:
            # Next gap interval is exist
            next_gap = new_gap_list[batch_index + 1]
            timeseries_train_part = output_data[(latest_gap_element_id + 1): next_gap[0]]

            # Take part with known values to the left from the gap
            extended_part = output_data[(first_gap_element_id - 1): next_gap[0]]

            if first_gap_element_id == 0:
                # Gap in the first part of time series - take first observed value
                first_known_value = timeseries_train_part[0]
                return weights_list, [first_known_value] * len_gap
        timeseries_train_part = np.flip(timeseries_train_part)

        train_ts_len = len(timeseries_train_part) - len_gap
        if train_ts_len < self.min_train_ts_length:
            interpolated_part = self.linear_interpolation(extended_part)
            # Clip pre-history
            interpolated_part = interpolated_part[1:]
            # Clip parts after gap interval
            predicted = interpolated_part[:len_gap]
        else:
            predicted = self.__pipeline_fit_predict(self.pipeline,
                                                    timeseries_train_part,
                                                    len_gap)

            predicted = np.flip(predicted)
        return weights_list, predicted

    def __pipeline_fit_predict(self, pipeline, timeseries_train: np.array, len_gap: int):
        """The method makes a prediction as a sequence of elements based on a
        training sample. There are two main parts: fit model and predict.

        Args:
            pipeline: pipeline for forecasting
            timeseries_train: part of the time series for training the model
            len_gap: number of elements in the gap

        Returns:
            array without gaps
        """
        pipeline_for_forecast = deepcopy(pipeline)

        task = Task(TaskTypesEnum.ts_forecasting,
                    TsForecastingParams(forecast_length=len_gap))

        input_data = InputData(idx=np.arange(0, len(timeseries_train)),
                               features=timeseries_train,
                               target=timeseries_train,
                               task=task,
                               data_type=DataTypesEnum.ts)

        # Making predictions for the missing part in the time series
        pipeline_for_forecast.fit_from_scratch(input_data)

        # "Test data" for making prediction for a specific length
        start_forecast = len(timeseries_train)
        end_forecast = start_forecast + len_gap
        idx_test = np.arange(start_forecast, end_forecast)
        test_data = InputData(idx=idx_test,
                              features=timeseries_train,
                              target=None,
                              task=task,
                              data_type=DataTypesEnum.ts)

        predicted_values = pipeline_for_forecast.predict(test_data)
        predicted_values = np.ravel(np.array(predicted_values.predict))
        return predicted_values

    def __forecast_in_gap(self, pipeline, timeseries_train_part, output_data, gap):
        """ Make forecast for desired part of time series with gap

        Args:
            pipeline: pipeline for forecasting
            timeseries_train_part: part of time series without gaps to fit pipeline
            output_data: array with gaps (some og them may be filled previously)
            gap: indices of continuous batch (gap)

        Returns:
            predicted values
        """

        train_ts_len = len(timeseries_train_part) - len(gap)
        if train_ts_len < self.min_train_ts_length:
            # Take part with gap [..., gap, gap, known_value]
            gap_part = output_data[:gap[-1] + 2]

            # Use linear interpolation - get full time series
            interpolated_part = self.linear_interpolation(gap_part)
            predicted = interpolated_part[gap]
        else:
            # Pipeline for the task of filling in gaps
            predicted = self.__pipeline_fit_predict(pipeline,
                                                    timeseries_train_part,
                                                    len(gap))

        return predicted


def replace_nan_with_label(time_series: np.ndarray, label: Union[int, float]):
    """ Replace np.nan in the array with desired label """
    return np.nan_to_num(time_series, nan=label)



import json
import os

import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error

from examples.advanced.time_series_forecasting.composing_pipelines import visualise, get_border_line_info
from examples.simple.pipeline_import_export import create_correct_path
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from fedot.core.utils import fedot_project_root


def get_ts_data_long(n_steps=80, forecast_length=5):
    """ Prepare data from csv file with time series and take needed number of
    elements

    :param n_steps: number of elements in time series to take
    :param forecast_length: the length of forecast
    """
    project_root_path = str(fedot_project_root())
    file_path = os.path.join(project_root_path, 'examples/data/ts/ts_long.csv')
    df = pd.read_csv(file_path)
    df = df[df["series_id"] == "traffic_volume"]
    time_series = np.array(df['value'])[:n_steps]
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=forecast_length))

    data = InputData(idx=np.arange(0, len(time_series)),
                     features=time_series,
                     target=time_series,
                     task=task,
                     data_type=DataTypesEnum.ts)
    return train_test_data_setup(data), task


def clstm_forecasting():
    """ Example of clstm pipeline serialization """
    horizon = 24 * 2
    window_size = 29
    n_steps = 100
    (train_data, test_data), _ = get_ts_data_long(n_steps=n_steps + horizon, forecast_length=horizon)

    node_root = PrimaryNode("clstm")
    node_root.parameters = {'window_size': window_size,
                            'hidden_size': 135,
                            'learning_rate': 0.0004,
                            'cnn1_kernel_size': 5,
                            'cnn1_output_size': 32,
                            'cnn2_kernel_size': 4,
                            'cnn2_output_size': 32,
                            'batch_size': 64,
                            'num_epochs': 50}

    pipeline = Pipeline(node_root)
    pipeline.fit(train_data)
    prediction_before_export = pipeline.predict(test_data).predict[0]

    print(f'Before export {prediction_before_export[:4]}')

    # Export it
    pipeline_path = "clstm"
    pipeline.save(path=pipeline_path)

    # Import pipeline
    json_path_load = create_correct_path(pipeline_path)
    new_pipeline = Pipeline.from_serialized(json_path_load)

    predicted_output_after_export = new_pipeline.predict(test_data)
    prediction_after_export = np.array(predicted_output_after_export.predict[0])

    print(f'After import {prediction_after_export[:4]}')

    dict_pipeline, dict_fitted_operations = pipeline.save()
    dict_pipeline = json.loads(dict_pipeline)
    pipeline_from_dict = Pipeline.from_serialized(dict_pipeline, dict_fitted_operations)

    predicted_output = pipeline_from_dict.predict(test_data)
    prediction = np.array(predicted_output.predict[0])
    print(f'Prediction from pipeline loaded from dict {prediction[:4]}')

    plot_info = [
        {'idx': np.arange(np.concatenate([test_data.features, test_data.target]).shape[0]),
         'series': np.concatenate([test_data.features, test_data.target]),
         'label': 'Actual time series'},
        {'idx': np.arange(test_data.idx[0], test_data.idx[0] + horizon),
         'series': np.ravel(prediction_before_export),
         'label': 'prediction'},
        get_border_line_info(np.arange(test_data.idx[0] + 1)[-1],
                             prediction,
                             np.ravel(np.concatenate([test_data.features, test_data.target])),
                             'Border line')]

    rmse = mean_squared_error(test_data.target, prediction_before_export, squared=False)
    mae = mean_absolute_error(test_data.target, prediction_before_export)
    print(f'RMSE - {rmse:.4f}')
    print(f'MAE - {mae:.4f}')

    visualise(plot_info)


if __name__ == '__main__':
    clstm_forecasting()



from abc import abstractmethod
from typing import Optional, Union, Dict, Any

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import default_log
from fedot.core.operations.hyperparameters_preprocessing import HyperparametersPreprocessor
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.operation_types_repository import OperationMetaInfo
from fedot.core.repository.tasks import Task, TaskTypesEnum, compatible_task_types
from fedot.core.serializers.serializer import register_serializable


@register_serializable
class Operation:
    """Base class for operations in nodes. Operations could be machine learning
    (or statistical) models or data operations

    Args:
        operation_type: name of the operation
    """

    def __init__(self, operation_type: str, **kwargs):
        self.operation_type = operation_type

        self._eval_strategy = None
        self.operations_repo = None
        self.fitted_operation = None

        self.log = default_log(self)

    def _init(self, task: Task, **kwargs):
        params = kwargs.get('params')
        if not params:
            params = OperationParameters.from_operation_type(self.operation_type)
        if isinstance(params, dict):
            params = OperationParameters.from_operation_type(self.operation_type, **params)
        params_for_fit = HyperparametersPreprocessor(operation_type=self.operation_type,
                                                     n_samples_data=kwargs.get('n_samples_data'))\
            .correct(params.to_dict())
        params_for_fit = OperationParameters.from_operation_type(self.operation_type, **params_for_fit)
        try:
            self._eval_strategy = \
                _eval_strategy_for_task(self.operation_type,
                                        task.task_type,
                                        self.operations_repo)(self.operation_type,
                                                              params_for_fit)
        except Exception as ex:
            self.log.error(f'Can not find evaluation strategy because of {ex}')
            raise ex

        if 'output_mode' in kwargs:
            self._eval_strategy.output_mode = kwargs['output_mode']

    def description(self, operation_params: dict) -> str:
        operation_type = self.operation_type
        return f'n_{operation_type}_{operation_params}'

    @property
    def acceptable_task_types(self):
        operation_info = self.operations_repo.operation_info_by_id(
            self.operation_type)
        return operation_info.task_type

    @property
    def metadata(self) -> OperationMetaInfo:
        operation_info = self.operations_repo.operation_info_by_id(self.operation_type)
        if not operation_info:
            raise ValueError(f'{self.__class__.__name__} {self.operation_type} not found')
        return operation_info

    def fit(self, params: Optional[Union[OperationParameters, dict]], data: InputData):
        """This method is used for defining and running of the evaluation strategy
        to train the operation with the data provided

        Args:
            params: hyperparameters for operation
            data: data used for operation training

        Returns:
            tuple: trained operation and prediction on train data
        """
        self._init(data.task, params=params, n_samples_data=data.features.shape[0])

        self.fitted_operation = self._eval_strategy.fit(train_data=data)

        predict_train = self.predict_for_fit(self.fitted_operation, data, params)

        return self.fitted_operation, predict_train

    def predict(self, fitted_operation, data: InputData, params: Optional[Union[OperationParameters, dict]] = None,
                output_mode: str = 'default'):
        """This method is used for defining and running of the evaluation strategy
        to predict with the data provided

        Args:
            fitted_operation: trained operation object
            data: data used for prediction
            params: hyperparameters for operation
            output_mode: string with information about output of operation,
            for example, is the operation predict probabilities or class labels
        """
        return self._predict(fitted_operation, data, params, output_mode, is_fit_stage=False)

    def predict_for_fit(self, fitted_operation, data: InputData, params: Optional[OperationParameters] = None,
                        output_mode: str = 'default'):
        """This method is used for defining and running of the evaluation strategy
        to predict with the data provided during fit stage

        Args:
            fitted_operation: trained operation object
            data: data used for prediction
            params: hyperparameters for operation
            output_mode: string with information about output of operation,
                for example, is the operation predict probabilities or class labels
        """
        return self._predict(fitted_operation, data, params, output_mode, is_fit_stage=True)

    def _predict(self, fitted_operation, data: InputData, params: Optional[OperationParameters] = None,
                 output_mode: str = 'default', is_fit_stage: bool = False):

        is_main_target = data.supplementary_data.is_main_target
        data_flow_length = data.supplementary_data.data_flow_length
        self._init(data.task, output_mode=output_mode, params=params, n_samples_data=data.features.shape[0])

        if is_fit_stage:
            prediction = self._eval_strategy.predict_for_fit(
                trained_operation=fitted_operation,
                predict_data=data)
        else:
            prediction = self._eval_strategy.predict(
                trained_operation=fitted_operation,
                predict_data=data)
        prediction = self.assign_tabular_column_types(prediction, output_mode)

        if is_main_target is False:
            prediction.supplementary_data.is_main_target = is_main_target

        prediction.supplementary_data.data_flow_length = data_flow_length
        prediction.supplementary_data.was_preprocessed = True
        return prediction

    @staticmethod
    @abstractmethod
    def assign_tabular_column_types(output_data: OutputData, output_mode: str) -> OutputData:
        """Assign types for columns based on task and output_mode (for classification)\n
        For example, pipeline for solving time series forecasting task contains lagged and ridge operations.
        ``ts_type -> lagged -> tabular type``\n
        So, there is a need to assign column types to new data
        """
        raise NotImplementedError()

    def __str__(self):
        return f'{self.operation_type}'

    def to_json(self) -> Dict[str, Any]:
        """Serializes object and ignores unrelevant fields."""
        return {
            k: v
            for k, v in sorted(vars(self).items())
            if k not in ['log', 'operations_repo', '_eval_strategy', 'fitted_operation']
        }


def _eval_strategy_for_task(operation_type: str, current_task_type: TaskTypesEnum,
                            operations_repo):
    """The function returns the strategy for the selected operation and task type.
    And if it is necessary, found acceptable strategy for operation

    Args:
        operation_type: name of operation, for example, ``'ridge'``
        current_task_type: task to solve
        operations_repo: repository with operations

    Returns:
        EvaluationStrategy: ``EvaluationStrategy`` class for this operation
    """

    # Get acceptable task types for operation
    operation_info = operations_repo.operation_info_by_id(operation_type)

    if operation_info is None:
        raise ValueError(f'{operation_type} is not implemented '
                         f'in {operations_repo.repository_name}')

    acceptable_task_types = operation_info.task_type

    # If the operation can't be used directly for the task type from data
    set_acceptable_types = set(acceptable_task_types)
    if current_task_type not in acceptable_task_types:

        # Search the supplementary task types, that can be included in pipeline
        # which solves main task
        globally_compatible_task_types = compatible_task_types(current_task_type)
        globally_set = set(globally_compatible_task_types)

        comp_types_acceptable_for_operation = list(set_acceptable_types.intersection(globally_set))
        if len(comp_types_acceptable_for_operation) == 0:
            raise ValueError(f'Operation {operation_type} can not be used as a part of {current_task_type}.')
        current_task_type = comp_types_acceptable_for_operation[0]

    strategy = operations_repo.operation_info_by_id(operation_type).current_strategy(current_task_type)
    return strategy



import os
import random

import numpy as np

from cases.credit_scoring.credit_scoring_problem import run_credit_scoring_problem, get_scoring_data
from fedot.utilities.profiler.memory_profiler import MemoryProfiler
from fedot.utilities.profiler.time_profiler import TimeProfiler

random.seed(1)
np.random.seed(1)

if __name__ == '__main__':
    # JUST UNCOMMENT WHAT TYPE OF PROFILER DO YOU NEED
    # EXAMPLE of MemoryProfiler.

    path = os.path.join(os.path.expanduser("~"), 'memory_profiler')
    full_path_train, full_path_test = get_scoring_data()
    arguments = {'train_file_path': full_path_train, 'test_file_path': full_path_test}
    MemoryProfiler(run_credit_scoring_problem, kwargs=arguments, path=path,
                   roots=[run_credit_scoring_problem], max_depth=8)

    # EXAMPLE of TimeProfiler.

    profiler = TimeProfiler()
    full_path_train, full_path_test = get_scoring_data()
    run_credit_scoring_problem(full_path_train, full_path_test)
    path = os.path.join(os.path.expanduser("~"), 'time_profiler')
    profiler.profile(path=path, node_percent=0.5, edge_percent=0.1, open_web=True)



import os
from typing import Optional

import numpy as np

from fedot.core.data.data import InputData, OutputData
from fedot.core.log import default_log
from fedot.core.operations.evaluation.operation_implementations. \
    implementation_interfaces import DataOperationImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.utilities.requirements_notificator import warn_requirement

try:
    import gensim.downloader as api
    from gensim.models import KeyedVectors
except ModuleNotFoundError:
    warn_requirement('gensim')
    api = None
    KeyedVectors = None


class PretrainedEmbeddingsImplementation(DataOperationImplementation):
    """ Class for text vectorization by pretrained embeddings
    model_name can be selected from https://github.com/RaRe-Technologies/gensim-data"""

    def __init__(self, params: Optional[OperationParameters]):
        super().__init__(params)
        self.logger = default_log(prefix='FEDOT logger')
        self._download_model_resources()

    def fit(self, input_data: InputData):
        """ Class doesn't support fit operation

        :param input_data: data with features, target and ids to process
        """
        pass

    def transform(self, input_data: InputData) -> OutputData:
        """ Method for transformation of the text data for predict stage

        :param input_data: data with features, target and ids to process
        :return output_data: output data with transformed features table
        """

        embed_data = np.stack([self.vectorize_avg(text, self.model) for text in input_data.features])
        output_data = self._convert_to_output(input_data,
                                              embed_data,
                                              data_type=DataTypesEnum.table)
        return output_data

    @staticmethod
    def vectorize_avg(text: str, embeddings):
        """ Method converts text to an average of token vectors

        :param text: str with text data
        :param embeddings: gensim pretrained embeddings
        :return features: one-dimensional np.array with numbers
        """
        embedding_dim = embeddings.vectors.shape[1]
        features = np.zeros([embedding_dim], dtype='float32')
        num_words = 0

        for word in text.split():
            if word in embeddings:
                features += embeddings[f'{word}']
                num_words += 1

        if num_words > 0:
            return features / num_words
        return features

    def _download_model_resources(self):
        """ Method for downloading text embeddings. Embeddings are loaded into external folder"""
        self.logger.info('Trying to download embeddings...')
        model_name = self.params.setdefault('model_name', 'glove-twitter-25')
        model_path = api.load(f"{model_name}", return_path=True)

        if os.path.exists(model_path):
            self.logger.info('Embeddings are already downloaded. Loading model...')
            self.model = KeyedVectors.load_word2vec_format(model_path, binary=False)



from collections import namedtuple
from typing import List, Optional

from fedot.core.pipelines.node import Node

HyperparamsAnalysisMetaParams = namedtuple('HyperparamsAnalysisMetaParams', ['analyze_method',
                                                                             'sample_method',
                                                                             'sample_size'])

ReplacementAnalysisMetaParams = namedtuple('ReplacementAnalysisMetaParams', ['nodes_to_replace_to',
                                                                             'number_of_random_operations'])


class SensitivityAnalysisRequirements:
    """Use this object to pass all the requirements needed for ``SA``

    Args:
        hyperparams_analyze_method: defines string name of SA method to use. Defaults: ``sobol``
        hyperparams_sample_method: defines string name of sampling method to use. Defaults: ``saltelli``
        hyperparams_analysis_samples_size: defines the number of shyperparameters samples used in SA
        replacement_nodes_to_replace_to: defines nodes which is used in replacement analysis.
        replacement_number_of_random_operations: if ``replacement_nodes_to_replace_to`` is not filled,
            define the number of randomly chosen operations used in replacement analysis.
        visualization: defines whether the SA visualization needs to be saved to ``.png`` files.
        is_save_results_to_json: defines whether the SA indices needs to be saved to ``.json`` file.
        metric: metric used for validation. Default: see :obj:`MetricByTask`
    """

    def __init__(self,
                 metric=None,
                 hyperparams_analyze_method: str = 'sobol',
                 hyperparams_sample_method: str = 'saltelli',
                 hyperparams_analysis_samples_size: int = 100,
                 replacement_nodes_to_replace_to: Optional[List[Node]] = None,
                 replacement_number_of_random_operations: Optional[int] = None,
                 visualization: bool = True,
                 is_save_results_to_json: bool = True):
        self.metric = metric
        self.hp_analysis_meta = HyperparamsAnalysisMetaParams(hyperparams_analyze_method,
                                                              hyperparams_sample_method,
                                                              hyperparams_analysis_samples_size,
                                                              )

        self.replacement_meta = ReplacementAnalysisMetaParams(replacement_nodes_to_replace_to,
                                                              replacement_number_of_random_operations)

        self.visualization = visualization
        self.is_save = is_save_results_to_json



from fedot.core.utilities.data_structures import ComparableEnum as Enum


class DataTypesEnum(Enum):
    """An enumeration

    Args:
        table: table with columns as features for predictions, by default == ``feature_table``
        ts: one dimensional array - time series, by default == ``time_series``
        multi_ts: table with different variant of time-series for the same variable as columns
           (used for extending train sample), by default == ``multiple_time_series``
        text: table, where cells contains text, by default == ``text``
        image: images represented as 3d arrays, by default == ``image``
    """

    # Table with columns as features for predictions
    table = 'feature_table'

    # One dimensional array - time series
    ts = 'time_series'

    # Table with different variant of time-series for the same variable as columns (used for extending train sample)
    multi_ts = 'multiple_time_series'

    # Table, where cells contains text
    text = 'text'

    # Images represented as 3d arrays
    image = 'image'
